import math
import os
import unittest

from dataclasses import dataclass
import numpy as np
import geopandas as gpd
import pandas as pd
from shapely.geometry import Polygon, JOIN_STYLE
from typing import List

from naturf.driver import Model
import naturf.nodes as nodes
from naturf.config import Settings


class TestNodes(unittest.TestCase):
    INPUTS = {
        "input_shapefile": os.path.join("naturf", "data", "C-5.shp"),
        "radius": 100,
        "cap_style": 1,
    }

    def test_area_weighted_mean_of_building_heights(self):
        "Test that the function `area_weighted_mean_of_building_heights()` returns the correct value."
        # ID 0: one building
        # ID 1: two buildings
        # ID 2: multiple buildings, decimal values
        data = [[0, 10, 10], [1, 2, 10], [1, 1, 10], [2, 0.1, 10], [2, 0.9, 30], [2, 12, 0.25]]
        buildings_intersecting_plan_area = pd.DataFrame(
            data,
            columns=[
                Settings.TARGET_ID_FIELD,
                Settings.NEIGHBOR_HEIGHT_FIELD,
                Settings.NEIGHBOR_AREA_FIELD,
            ],
        )
        expected = pd.Series([10.0, 1.5, 0.7701863354037267])
        actual = nodes.area_weighted_mean_of_building_heights(buildings_intersecting_plan_area)
        pd.testing.assert_series_equal(
            expected,
            actual,
            f"area_weighted_mean_of_building_heights test failed, expected {expected}, actual {actual}",
        )

    def test_average_distance_between_buildings(self):
        "Test that the function `average_distance_between_buildings()` returns the correct distance."

        # Each ID number refers to a particular building and test case.
        # Buildings 0, 1, and 2 test that the mean function is working correctly.
        # Building 3 tests that a distance of zero (representing a building considered its own neighbor) does not affect the mean.
        # Building 4 tests that a building with only itself as a neighbor returns the default street width.

        index = pd.Index([0, 0, 1, 1, 2, 2, 3, 3, 3, 4], name=Settings.NEIGHBOR_ID_FIELD)
        distance = pd.Series(
            [1, 2, 1, 3, 1, 1, 3, 0, 3, 0], index, name=Settings.DISTANCE_BETWEEN_BUILDINGS
        )

        default_street_width = Settings.DEFAULT_STREET_WIDTH

        expected = pd.Series(
            [1.5, 2.0, 1.0, 3.0, default_street_width],
            name=Settings.AVERAGE_DISTANCE_BETWEEN_BUILDINGS,
        )

        actual = nodes.average_distance_between_buildings(distance)

        pd.testing.assert_series_equal(
            expected,
            actual,
            f"average_distance_between_buildings test failed, expected {expected}, actual {actual}",
        )

    def test_buildings_intersecting_plan_area(self):
        """Test that the function `buildings_intersecting_plan_area()` returns the correct intersecting buildings."""

        polygon1 = Polygon([[0, 0], [0, 1], [1, 1], [1, 0]])
        polygon2 = Polygon([[3, 3], [3, 4], [4, 4], [4, 3]])
        building_id = pd.Series([0, 1])
        building_height = pd.Series([5, 10])
        building_geometry = pd.Series([polygon1, polygon2])
        building_area = pd.Series([polygon1.area, polygon2.area])
        crs = "epsg:3857"

        wall_length_north = pd.Series([1.0, 1.0], name=Settings.WALL_LENGTH_NORTH)
        wall_length_east = pd.Series([1.0, 1.0], name=Settings.WALL_LENGTH_EAST)
        wall_length_south = pd.Series([1.0, 1.0], name=Settings.WALL_LENGTH_SOUTH)
        wall_length_west = pd.Series([1.0, 1.0], name=Settings.WALL_LENGTH_WEST)

        wall_length = pd.concat(
            [
                wall_length_north,
                wall_length_east,
                wall_length_south,
                wall_length_west,
            ],
            axis=1,
        )

        total_plan_area_geometry_no_overlap = pd.Series([polygon1.buffer(1), polygon2.buffer(1)])
        total_plan_area_geometry_some_overlap = pd.Series(
            [polygon1.buffer(3.5), polygon2.buffer(3.5)]
        )
        total_plan_area_geometry_total_overlap = pd.Series([polygon1.buffer(5), polygon2.buffer(5)])

        no_overlap_output_gdf = gpd.GeoDataFrame(
            {
                "building_id_neighbor": building_id,
                "building_id_target": building_id,
                "building_height_target": building_height,
                "building_area_target": building_area,
                "building_geometry": building_geometry,
                "building_buffered_target": gpd.GeoSeries(total_plan_area_geometry_no_overlap),
                "wall_length_north_target": wall_length_north,
                "wall_length_east_target": wall_length_east,
                "wall_length_south_target": wall_length_south,
                "wall_length_west_target": wall_length_west,
                "index_neighbor": building_id,
                "building_height_neighbor": building_height,
                "building_area_neighbor": building_area,
                "building_buffered_neighbor": total_plan_area_geometry_no_overlap,
                "wall_length_north_neighbor": wall_length_north,
                "wall_length_east_neighbor": wall_length_east,
                "wall_length_south_neighbor": wall_length_south,
                "wall_length_west_neighbor": wall_length_west,
                "building_geometry_neighbor": gpd.GeoSeries(building_geometry),
            },
            geometry="building_geometry",
        ).set_index("building_id_neighbor")

        some_overlap_output_gdf = gpd.GeoDataFrame(
            {
                "building_id_neighbor": pd.Series([0, 0, 1, 1]),
                "building_id_target": pd.Series([0, 1, 0, 1]),
                "building_height_target": pd.Series([5, 10, 5, 10]),
                "building_area_target": pd.Series([1.0, 1.0, 1.0, 1.0]),
                "building_geometry": gpd.GeoSeries(
                    [
                        building_geometry[0],
                        building_geometry[1],
                        building_geometry[0],
                        building_geometry[1],
                    ]
                ),
                "building_buffered_target": gpd.GeoSeries(
                    [
                        total_plan_area_geometry_some_overlap[0],
                        total_plan_area_geometry_some_overlap[1],
                        total_plan_area_geometry_some_overlap[0],
                        total_plan_area_geometry_some_overlap[1],
                    ]
                ),
                "wall_length_north_target": pd.Series([1.0, 1.0, 1.0, 1.0]),
                "wall_length_east_target": pd.Series([1.0, 1.0, 1.0, 1.0]),
                "wall_length_south_target": pd.Series([1.0, 1.0, 1.0, 1.0]),
                "wall_length_west_target": pd.Series([1.0, 1.0, 1.0, 1.0]),
                "index_neighbor": pd.Series([0, 0, 1, 1]),
                "building_height_neighbor": pd.Series([5, 5, 10, 10]),
                "building_area_neighbor": pd.Series([1.0, 1.0, 1.0, 1.0]),
                "building_buffered_neighbor": pd.Series(
                    [
                        total_plan_area_geometry_some_overlap[0],
                        total_plan_area_geometry_some_overlap[0],
                        total_plan_area_geometry_some_overlap[1],
                        total_plan_area_geometry_some_overlap[1],
                    ]
                ),
                "wall_length_north_neighbor": pd.Series([1.0, 1.0, 1.0, 1.0]),
                "wall_length_east_neighbor": pd.Series([1.0, 1.0, 1.0, 1.0]),
                "wall_length_south_neighbor": pd.Series([1.0, 1.0, 1.0, 1.0]),
                "wall_length_west_neighbor": pd.Series([1.0, 1.0, 1.0, 1.0]),
                "building_geometry_neighbor": gpd.GeoSeries(
                    [
                        building_geometry[0],
                        building_geometry[0],
                        building_geometry[1],
                        building_geometry[1],
                    ]
                ),
            },
            geometry="building_geometry",
        ).set_index("building_id_neighbor")

        total_overlap_output_gdf = gpd.GeoDataFrame(
            {
                "building_id_neighbor": pd.Series([0, 0, 1, 1]),
                "building_id_target": pd.Series([0, 1, 0, 1]),
                "building_height_target": pd.Series([5, 10, 5, 10]),
                "building_area_target": pd.Series([1.0, 1.0, 1.0, 1.0]),
                "building_geometry": gpd.GeoSeries(
                    [
                        building_geometry[0],
                        building_geometry[1],
                        building_geometry[0],
                        building_geometry[1],
                    ]
                ),
                "building_buffered_target": gpd.GeoSeries(
                    [
                        total_plan_area_geometry_total_overlap[0],
                        total_plan_area_geometry_total_overlap[1],
                        total_plan_area_geometry_total_overlap[0],
                        total_plan_area_geometry_total_overlap[1],
                    ]
                ),
                "wall_length_north_target": pd.Series([1.0, 1.0, 1.0, 1.0]),
                "wall_length_east_target": pd.Series([1.0, 1.0, 1.0, 1.0]),
                "wall_length_south_target": pd.Series([1.0, 1.0, 1.0, 1.0]),
                "wall_length_west_target": pd.Series([1.0, 1.0, 1.0, 1.0]),
                "index_neighbor": pd.Series([0, 0, 1, 1]),
                "building_height_neighbor": pd.Series([5, 5, 10, 10]),
                "building_area_neighbor": pd.Series([1.0, 1.0, 1.0, 1.0]),
                "building_buffered_neighbor": pd.Series(
                    [
                        total_plan_area_geometry_total_overlap[0],
                        total_plan_area_geometry_total_overlap[0],
                        total_plan_area_geometry_total_overlap[1],
                        total_plan_area_geometry_total_overlap[1],
                    ]
                ),
                "wall_length_north_neighbor": pd.Series([1.0, 1.0, 1.0, 1.0]),
                "wall_length_east_neighbor": pd.Series([1.0, 1.0, 1.0, 1.0]),
                "wall_length_south_neighbor": pd.Series([1.0, 1.0, 1.0, 1.0]),
                "wall_length_west_neighbor": pd.Series([1.0, 1.0, 1.0, 1.0]),
                "building_geometry_neighbor": gpd.GeoSeries(
                    [
                        building_geometry[0],
                        building_geometry[0],
                        building_geometry[1],
                        building_geometry[1],
                    ]
                ),
            },
            geometry="building_geometry",
        ).set_index("building_id_neighbor")

        @dataclass
        class TestCase:
            name: str
            input: pd.Series
            expected: gpd.GeoDataFrame

        testcases = [
            TestCase(
                name="no overlap",
                input=total_plan_area_geometry_no_overlap,
                expected=no_overlap_output_gdf,
            ),
            TestCase(
                name="some overlap",
                input=total_plan_area_geometry_some_overlap,
                expected=some_overlap_output_gdf,
            ),
            TestCase(
                name="total overlap",
                input=total_plan_area_geometry_total_overlap,
                expected=total_overlap_output_gdf,
            ),
        ]

        for case in testcases:
            actual = nodes.buildings_intersecting_plan_area(
                building_id,
                building_height,
                building_geometry,
                building_area,
                case.input,
                wall_length,
                crs,
            )
            expected = case.expected
            pd.testing.assert_frame_equal(
                expected,
                actual,
                f"buildings_intersecting_plan_area test {case.name} failed, expected {expected}, actual {actual}",
            )

    def test_building_plan_area(self):
        """Test that the function `building_plan_area()` returns the correct area."""

        polygon1 = Polygon([[0, 0], [0, 1], [1, 1], [1, 0]])
        polygon2 = Polygon([[3, 3], [3, 4], [4, 4], [4, 3]])
        building_id = pd.Series([0, 1])
        building_height = pd.Series([5, 10])
        building_geometry = pd.Series([polygon1, polygon2])
        building_area = pd.Series([polygon1.area, polygon2.area])

        total_plan_area_geometry_no_overlap = pd.Series([polygon1.buffer(1), polygon2.buffer(1)])
        total_plan_area_geometry_some_overlap = pd.Series(
            [
                polygon1.buffer(2.5, join_style=JOIN_STYLE.mitre),
                polygon2.buffer(2.5, join_style=JOIN_STYLE.mitre),
            ]
        )
        total_plan_area_geometry_total_overlap = pd.Series([polygon1.buffer(5), polygon2.buffer(5)])

        no_overlap_gdf = gpd.GeoDataFrame(
            {
                "building_id_neighbor": building_id,
                "building_id_target": building_id,
                "building_height_target": building_height,
                "building_area_target": building_area,
                "building_geometry": building_geometry,
                "building_buffered_target": gpd.GeoSeries(total_plan_area_geometry_no_overlap),
                "index_neighbor": building_id,
                "building_height_neighbor": building_height,
                "building_area_neighbor": building_area,
                "building_buffered_neighbor": total_plan_area_geometry_no_overlap,
                "building_geometry_neighbor": gpd.GeoSeries(building_geometry),
            },
            geometry="building_geometry",
        ).set_index("building_id_neighbor")

        some_overlap_gdf = gpd.GeoDataFrame(
            {
                "building_id_neighbor": pd.Series([0, 0, 1, 1]),
                "building_id_target": pd.Series([0, 1, 0, 1]),
                "building_height_target": pd.Series([5, 10, 5, 10]),
                "building_area_target": pd.Series([1.0, 1.0, 1.0, 1.0]),
                "building_geometry": gpd.GeoSeries(
                    [
                        building_geometry[0],
                        building_geometry[1],
                        building_geometry[0],
                        building_geometry[1],
                    ]
                ),
                "building_buffered_target": gpd.GeoSeries(
                    [
                        total_plan_area_geometry_some_overlap[0],
                        total_plan_area_geometry_some_overlap[1],
                        total_plan_area_geometry_some_overlap[0],
                        total_plan_area_geometry_some_overlap[1],
                    ]
                ),
                "index_neighbor": pd.Series([0, 0, 1, 1]),
                "building_height_neighbor": pd.Series([5, 5, 10, 10]),
                "building_area_neighbor": pd.Series([1.0, 1.0, 1.0, 1.0]),
                "building_buffered_neighbor": pd.Series(
                    [
                        total_plan_area_geometry_some_overlap[0],
                        total_plan_area_geometry_some_overlap[0],
                        total_plan_area_geometry_some_overlap[1],
                        total_plan_area_geometry_some_overlap[1],
                    ]
                ),
                "building_geometry_neighbor": gpd.GeoSeries(
                    [
                        building_geometry[0],
                        building_geometry[0],
                        building_geometry[1],
                        building_geometry[1],
                    ]
                ),
            },
            geometry="building_geometry",
        ).set_index("building_id_neighbor")

        total_overlap_gdf = gpd.GeoDataFrame(
            {
                "building_id_neighbor": pd.Series([0, 0, 1, 1]),
                "building_id_target": pd.Series([0, 1, 0, 1]),
                "building_height_target": pd.Series([5, 10, 5, 10]),
                "building_area_target": pd.Series([1.0, 1.0, 1.0, 1.0]),
                "building_geometry": gpd.GeoSeries(
                    [
                        building_geometry[0],
                        building_geometry[1],
                        building_geometry[0],
                        building_geometry[1],
                    ]
                ),
                "building_buffered_target": gpd.GeoSeries(
                    [
                        total_plan_area_geometry_total_overlap[0],
                        total_plan_area_geometry_total_overlap[1],
                        total_plan_area_geometry_total_overlap[0],
                        total_plan_area_geometry_total_overlap[1],
                    ]
                ),
                "index_neighbor": pd.Series([0, 0, 1, 1]),
                "building_height_neighbor": pd.Series([5, 5, 10, 10]),
                "building_area_neighbor": pd.Series([1.0, 1.0, 1.0, 1.0]),
                "building_buffered_neighbor": pd.Series(
                    [
                        total_plan_area_geometry_total_overlap[0],
                        total_plan_area_geometry_total_overlap[0],
                        total_plan_area_geometry_total_overlap[1],
                        total_plan_area_geometry_total_overlap[1],
                    ]
                ),
                "building_geometry_neighbor": gpd.GeoSeries(
                    [
                        building_geometry[0],
                        building_geometry[0],
                        building_geometry[1],
                        building_geometry[1],
                    ]
                ),
            },
            geometry="building_geometry",
        ).set_index("building_id_neighbor")

        @dataclass
        class TestCase:
            name: str
            input: gpd.GeoDataFrame
            expected: List[float]

        testcases = [
            TestCase(name="no overlap", input=no_overlap_gdf, expected=[1.0, 1.0]),
            TestCase(name="some overlap", input=some_overlap_gdf, expected=[1.25, 1.25]),
            TestCase(name="total overlap", input=total_overlap_gdf, expected=[2.0, 2.0]),
        ]

        for case in testcases:
            actual = nodes.building_plan_area(case.input)
            expected = pd.Series(case.expected)
            pd.testing.assert_series_equal(
                expected,
                actual,
                f"building_plan_area test {case.name} failed, expected {expected}, actual {actual}",
            )

    def test_building_surface_area_to_plan_area_ratio(self):
        """Test that the function `building_surface_area_to_plan_area_ratio()` returns the correct value."""

        building_surface_area = pd.Series([0, 1.5, 75, 0, 1.5, 75])
        total_plan_area = pd.Series([0, 0, 1.5, 1.5, 75, 75])
        expected = pd.Series([math.nan, math.inf, 50.0, 0.0, 0.02, 1.0])
        actual = nodes.building_surface_area_to_plan_area_ratio(
            building_surface_area, total_plan_area
        )
        pd.testing.assert_series_equal(
            expected,
            actual,
            f"building_surface_area_to_plan_area_ratio test failed, expected {expected}, actual {actual}",
        )

    def test_complete_aspect_ratio(self):
        """Test that the function `complete_aspect_ratio()` returns the correct value."""

        building_surface_area = pd.Series([0, 0, 0, 1, 1, 1, 5.5, 5.5, 5.5, 75, 75, 75])
        total_plan_area = pd.Series([0, 1, 5.5, 75, 0, 1, 5.5, 75, 0, 1, 5.5, 75])
        building_plan_area = pd.Series([0, 0, 1, 1, 5.5, 5.5, 75, 75, 0, 1, 5.5, 75])
        expected = pd.Series(
            [
                float("nan"),
                1.0,
                0.8181818181818182,
                1.0,
                -float("inf"),
                -3.5,
                -11.636363636363637,
                0.07333333333333333,
                float("inf"),
                75.0,
                13.636363636363637,
                1.0,
            ]
        )
        actual = nodes.complete_aspect_ratio(
            building_surface_area, total_plan_area, building_plan_area
        )
        pd.testing.assert_series_equal(
            expected,
            actual,
            f"complete_aspect_ratio test failed, expected {expected}, actual {actual}",
        )

    def test_input_shapefile_df(self):
        """Test that the function `input_shapefile_df()` creates the right shape and type of DataFrame."""

        # instantiate DAG asking for the output of input_shapefile_df()
        dag = Model(inputs=TestNodes.INPUTS, outputs=["input_shapefile_df"])

        # generate the output data frame from the driver
        df = dag.execute()

        # check shape of data frame
        self.assertEqual((260, 3), df.shape, "`input_shapefile_df` shape does not match expected")

        # check data types
        fake_geodataframe = gpd.GeoDataFrame(
            {
                "a": np.array([], dtype=np.int64),
                "b": np.array([], dtype=np.float64),
                "geometry": np.array([], dtype=gpd.array.GeometryDtype),
            }
        )

        np.testing.assert_array_equal(
            fake_geodataframe.dtypes.values,
            df.dtypes.values,
            "`input_shapefile_df` column data types do not match expected",
        )

        self.assertEqual(
            type(fake_geodataframe),
            type(df),
            "`input_shapefile_df` doesn't match expected data type.",
        )

    def test_frontal_area(self):
        """Test that the function `frontal_area()` returns the correct values."""
        frontal_length = pd.DataFrame([[0, 0.25, 0.5, 7500], [0, 3.14159, 10.5, 100]])
        building_height = pd.Series([0.5, 75])
        expected = pd.DataFrame([[0.0, 0.125, 0.25, 3750.0], [0.0, 235.61925, 787.5, 7500.0]])
        actual = nodes.frontal_area(frontal_length, building_height)
        expected.columns = [
            Settings.FRONTAL_AREA_NORTH,
            Settings.FRONTAL_AREA_EAST,
            Settings.FRONTAL_AREA_SOUTH,
            Settings.FRONTAL_AREA_WEST,
        ]
        pd.testing.assert_frame_equal(
            expected,
            actual,
            f"frontal_area test failed, expected {expected}, actual {actual}",
        )

    def test_frontal_area_density(self):
        """Test that the function `frontal_area_density()` returns the correct value."""

        # This uses different heights to test the function.
        # Building 1 is 5m tall to test that using one height bin works correctly.
        # Building 2 is 4m tall to test that using a height less than a full bin works correctly.
        # Building 3 is 21m tall to test that using a height >5m but less than a full bin works correctly.
        # Building 4 is 75m tall to test that a full-height building works correctly.

        frontal_length_north = Settings.FRONTAL_LENGTH_NORTH
        frontal_length_east = Settings.FRONTAL_LENGTH_EAST
        frontal_length_south = Settings.FRONTAL_LENGTH_SOUTH
        frontal_length_west = Settings.FRONTAL_LENGTH_WEST

        frontal_lengths = pd.concat(
            [
                pd.Series([2.0, 2.0, 2.0, 2.0], name=frontal_length_north),
                pd.Series([4.0, 4.0, 4.0, 4.0], name=frontal_length_east),
                pd.Series([1.0, 1.0, 1.0, 1.0], name=frontal_length_south),
                pd.Series([3.0, 3.0, 3.0, 3.0], name=frontal_length_west),
            ],
            axis=1,
        )
        heights = pd.Series([5, 4, 21, 75])

        total_plan_area = pd.Series([100.0, 100.0, 100.0, 100.0])

        rows = 4
        cols = 15

        frontal_area_north, frontal_area_east, frontal_area_south, frontal_area_west = (
            [[0 for i in range(cols)] for j in range(rows)],
            [[0 for i in range(cols)] for j in range(rows)],
            [[0 for i in range(cols)] for j in range(rows)],
            [[0 for i in range(cols)] for j in range(rows)],
        )

        frontal_area_north[0][0] = 0.1
        frontal_area_north[0][1] = 0.0

        frontal_area_east[0][0] = 0.2
        frontal_area_east[0][1] = 0.0

        frontal_area_south[0][0] = 0.05
        frontal_area_south[0][1] = 0.0

        frontal_area_west[0][0] = 0.15
        frontal_area_west[0][1] = 0.0

        frontal_area_north[1][0] = 0.08
        frontal_area_north[1][1] = 0.0

        frontal_area_east[1][0] = 0.16
        frontal_area_east[1][1] = 0.0

        frontal_area_south[1][0] = 0.04
        frontal_area_south[1][1] = 0.0

        frontal_area_west[1][0] = 0.12
        frontal_area_west[1][1] = 0.0

        for i in range(0, 4):
            frontal_area_north[2][i] = 0.1
        frontal_area_north[2][4] = 0.02

        for i in range(0, 4):
            frontal_area_east[2][i] = 0.20
        frontal_area_east[2][4] = 0.04

        for i in range(0, 4):
            frontal_area_south[2][i] = 0.05
        frontal_area_south[2][4] = 0.01

        for i in range(0, 4):
            frontal_area_west[2][i] = 0.15
        frontal_area_west[2][4] = 0.03

        for i in range(0, 15):
            frontal_area_north[3][i] = 0.1

        for i in range(0, 15):
            frontal_area_east[3][i] = 0.2

        for i in range(0, 15):
            frontal_area_south[3][i] = 0.05

        for i in range(0, 15):
            frontal_area_west[3][i] = 0.15

        columns_north, columns_east, columns_south, columns_west = (
            [
                f"{Settings.FRONTAL_AREA_NORTH}_{i}"
                for i in range(
                    int(Settings.MAX_BUILDING_HEIGHT / Settings.BUILDING_HEIGHT_INTERVAL)
                )
            ],
            [
                f"{Settings.FRONTAL_AREA_EAST}_{i}"
                for i in range(
                    int(Settings.MAX_BUILDING_HEIGHT / Settings.BUILDING_HEIGHT_INTERVAL)
                )
            ],
            [
                f"{Settings.FRONTAL_AREA_SOUTH}_{i}"
                for i in range(
                    int(Settings.MAX_BUILDING_HEIGHT / Settings.BUILDING_HEIGHT_INTERVAL)
                )
            ],
            [
                f"{Settings.FRONTAL_AREA_WEST}_{i}"
                for i in range(
                    int(Settings.MAX_BUILDING_HEIGHT / Settings.BUILDING_HEIGHT_INTERVAL)
                )
            ],
        )

        actual = nodes.frontal_area_density(frontal_lengths, heights, total_plan_area)
        expected = pd.concat(
            [
                pd.DataFrame(frontal_area_north, columns=columns_north),
                pd.DataFrame(frontal_area_east, columns=columns_east),
                pd.DataFrame(frontal_area_south, columns=columns_south),
                pd.DataFrame(frontal_area_west, columns=columns_west),
            ],
            axis=1,
        )
        # f"frontal_area_density test failed, expected {expected}, actual {actual}",
        pd.testing.assert_frame_equal(
            expected,
            actual,
        )

    def test_frontal_area_index(self):
        """Test that the function `frontal_area_index()` returns the correct value in each cardinal direction."""

        frontal_area = pd.DataFrame(
            [
                [0, 1, 5.5, 75],
                [0, 1, 5.5, 75],
                [0, 1, 5.5, 75],
            ]
        )
        total_plan_area = pd.Series([0, 1.5, 75])
        expected = pd.DataFrame(
            [
                [math.nan, math.inf, math.inf, math.inf],
                [0.0, 0.6666666666666666, 3.6666666666666665, 50.0],
                [0.0, 0.013333333333333334, 0.07333333333333333, 1.0],
            ]
        )
        expected.columns = [
            Settings.FRONTAL_AREA_INDEX_NORTH,
            Settings.FRONTAL_AREA_INDEX_EAST,
            Settings.FRONTAL_AREA_INDEX_SOUTH,
            Settings.FRONTAL_AREA_INDEX_WEST,
        ]
        actual = nodes.frontal_area_index(frontal_area, total_plan_area)
        # f"frontal_area_index test failed, expected {expected}, actual {actual}",
        pd.testing.assert_frame_equal(
            expected,
            actual,
        )

    def test_frontal_length(self):
        """Test that the function `frontal_length()` returns the correct length."""

        polygon1 = Polygon([[0, 0], [0, 1], [1, 1], [1, 0]])
        polygon2 = Polygon([[3, 3], [3, 4], [4, 4], [4, 3]])
        building_id = pd.Series([0, 1])
        building_height = pd.Series([5, 10])
        building_geometry = pd.Series([polygon1, polygon2])
        building_area = pd.Series([polygon1.area, polygon2.area])
        crs = "epsg:3857"

        # The wall lengths listed below are the test cases.
        # wall_length_north and wall_length_east test that the addition is working correctly.
        # wall_length_south tests that addition works correctly when one building does not have a wall facing a given direction.
        # wall_length_west tests that addition works correctly when neither building has a wall facing a given direction.

        wall_length_north = pd.Series([1.0, 1.0], name=Settings.WALL_LENGTH_NORTH)
        wall_length_east = pd.Series([1.0, 3.0], name=Settings.WALL_LENGTH_EAST)
        wall_length_south = pd.Series([0, 1.0], name=Settings.WALL_LENGTH_SOUTH)
        wall_length_west = pd.Series([0, 0], name=Settings.WALL_LENGTH_WEST)

        wall_length = pd.concat(
            [
                wall_length_north,
                wall_length_east,
                wall_length_south,
                wall_length_west,
            ],
            axis=1,
        )

        frontal_length_north = Settings.FRONTAL_LENGTH_NORTH
        frontal_length_east = Settings.FRONTAL_LENGTH_EAST
        frontal_length_south = Settings.FRONTAL_LENGTH_SOUTH
        frontal_length_west = Settings.FRONTAL_LENGTH_WEST

        total_plan_area_geometry = pd.Series([polygon1.buffer(5), polygon2.buffer(5)])

        input_gdf = nodes.buildings_intersecting_plan_area(
            building_id,
            building_height,
            building_geometry,
            building_area,
            total_plan_area_geometry,
            wall_length,
            crs,
        )

        actual = nodes.frontal_length(input_gdf)
        expected = pd.concat(
            [
                pd.Series([2.0, 2.0], name=frontal_length_north),
                pd.Series([4.0, 4.0], name=frontal_length_east),
                pd.Series([1.0, 1.0], name=frontal_length_south),
                pd.Series([0, 0], name=frontal_length_west),
            ],
            axis=1,
        )
        # f"frontal_length test failed, expected {expected}, actual {actual}"
        pd.testing.assert_frame_equal(
            expected,
            actual,
        )

    def test_grimmond_oke_displacement_height(self):
        """Test that the function `grimmond_oke_displacement_height()` returns the correct value."""
        building_height = pd.Series([0, 1, 10.5, 75])
        expected = pd.Series([0.0, 0.67, 7.035, 50.25])
        actual = nodes.grimmond_oke_displacement_height(building_height)
        # f"grimmond_oke_displacement_height test failed, expected {expected}, actual {actual}",
        pd.testing.assert_series_equal(
            expected,
            actual,
        )

    def test_grimmond_oke_roughness_length(self):
        """Test that the function `grimmond_oke_roughness_length()` returns the correct value."""
        building_height = pd.Series([0, 1, 10.5, 75])
        expected = pd.Series([0.0, 0.1, 1.05, 7.5])
        actual = nodes.grimmond_oke_roughness_length(building_height)
        # f"grimmond_oke_roughness_length test failed, expected {expected}, actual {actual}",
        pd.testing.assert_series_equal(
            expected,
            actual,
        )

    def test_height_to_width_ratio(self):
        """Test that the function `height_to_width_ratio()` returns the correct value."""
        mean_building_height = pd.Series([0, 0, 1, 10.5, 75])
        average_distance_between_buildings = pd.Series([0, 1, 0, 10.5, 100])
        expected = pd.Series([float("nan"), 0.0, float("inf"), 1.0, 0.75])
        actual = nodes.height_to_width_ratio(
            mean_building_height, average_distance_between_buildings
        )
        # f"height_to_width_ratio test failed, expected {expected}, actual {actual}",
        pd.testing.assert_series_equal(
            expected,
            actual,
        )

    def test_lot_area(self):
        """Test that the function `lot_area()` returns the correct values."""

        buildings_intersecting_plan_area = gpd.GeoDataFrame(
            [[0, 0], [0, 1], [1, 0], [0, 2], [2, 2], [1, 1]]
        )
        buildings_intersecting_plan_area.columns = [
            Settings.TARGET_ID_FIELD,
            f"index_{Settings.NEIGHBOR}",
        ]
        building_surface_area = pd.Series([1, 10, 100])

        expected = pd.Series([37.0, 5.5, 100.0])
        actual = nodes.lot_area(buildings_intersecting_plan_area, building_surface_area)
        # f"lot_area test failed, expected {expected}, actual {actual}",
        pd.testing.assert_series_equal(
            expected,
            actual,
        )

    def test_macdonald_displacement_height(self):
        """Test that the function `macdonald_displacement_height()` returns the correct height."""
        alpha_coefficient = Settings.ALPHACOEFFICIENT
        building_height = pd.Series([0, 10, 10, 10, 75], copy=False)
        plan_area_fraction = pd.Series([0.5, 0.5, 0, 1, 0.5], copy=False)
        expected = pd.Series(
            [
                0,
                10 - (5 / math.pow(alpha_coefficient, 0.5)),
                0,
                10,
                75 - (37.5 / math.pow(alpha_coefficient, 0.5)),
            ],
            copy=False,
        )
        actual = nodes.macdonald_displacement_height(building_height, plan_area_fraction)
        # f"macdonald_displacement_height test failed, expected {expected}, actual {actual}",
        pd.testing.assert_series_equal(
            expected,
            actual,
        )

    def test_macdonald_roughness_length(self):
        """Test that the function `macdonald_roughness_length()` returns the correct values."""
        building_height = pd.Series([0, 10.5, 75])
        macdonald_displacement_height = pd.Series([1, 0.5, 0])
        frontal_area = pd.DataFrame(
            [[0, 1, 1.5, 75], [0, 1, 1.5, 75], [0, 1, 1.5, 75]],
            columns=[
                Settings.FRONTAL_AREA_INDEX_NORTH,
                Settings.FRONTAL_AREA_INDEX_EAST,
                Settings.FRONTAL_AREA_INDEX_SOUTH,
                Settings.FRONTAL_AREA_INDEX_WEST,
            ],
        )
        lot_area = pd.Series([10.5, 1, 1])
        expected = pd.DataFrame(
            [
                [math.nan, math.nan, math.nan, math.nan],
                [0.0, 5.7826527777812675, 6.39407319161897, 9.387129414165152],
                [0.0, 43.94617676155573, 48.47520029599251, 70.51086232991824],
            ]
        )
        expected.columns = [
            Settings.MACDONALD_ROUGHNESS_LENGTH_NORTH,
            Settings.MACDONALD_ROUGHNESS_LENGTH_EAST,
            Settings.MACDONALD_ROUGHNESS_LENGTH_SOUTH,
            Settings.MACDONALD_ROUGHNESS_LENGTH_WEST,
        ]
        actual = nodes.macdonald_roughness_length(
            building_height, macdonald_displacement_height, frontal_area, lot_area
        )
        # f"macdonald_displacement_height test failed, expected {expected}, actual {actual}",
        pd.testing.assert_frame_equal(
            expected,
            actual,
        )

    def test_mean_building_height(self):
        """Test that the function `mean_building_height()` returns the correct values."""

        buildings_intersecting_plan_area = gpd.GeoDataFrame(
            [[0, 10], [0, 5], [1, 0], [2, 3.14], [2, 11358], [0, 15]]
        )
        buildings_intersecting_plan_area.columns = [
            Settings.TARGET_ID_FIELD,
            Settings.NEIGHBOR_HEIGHT_FIELD,
        ]
        expected = pd.Series([10, 0.0, 5680.57])
        actual = nodes.mean_building_height(buildings_intersecting_plan_area)
        # f"mean_building_height test failed, expected {expected}, actual {actual}",
        pd.testing.assert_series_equal(
            expected,
            actual,
        )

    def test_plan_area_density(self):
        """Test that the function `plan_area_density()` returns the correct value."""

        # This uses different heights to test the function.
        # Plan area density should be the same whether the building height falls at or within the bins, which all buildings test.
        # Height bins that are greater than the building height should have a plan area density of zero.

        building_plan_area = pd.Series([20, 25, 30, 35])
        building_height = pd.Series([5, 4, 21, 75])
        total_plan_area = pd.Series([100, 100, 100, 100])

        rows, cols = (
            len(building_height.index),
            int(Settings.MAX_BUILDING_HEIGHT / Settings.BUILDING_HEIGHT_INTERVAL),
        )
        plan_area_density = [[0 for i in range(cols)] for j in range(rows)]

        plan_area_density[0][0] = 0.2

        plan_area_density[1][0] = 0.25

        for i in range(0, 4):
            plan_area_density[2][i] = 0.3
        plan_area_density[2][4] = 0.3

        for i in range(0, 15):
            plan_area_density[3][i] = 0.35

        columns_plan_area_density = [
            f"{Settings.PLAN_AREA_DENSITY}_{i}"
            for i in range(int(Settings.MAX_BUILDING_HEIGHT / Settings.BUILDING_HEIGHT_INTERVAL))
        ]

        actual = nodes.plan_area_density(building_plan_area, building_height, total_plan_area)
        expected = pd.DataFrame(plan_area_density, columns=columns_plan_area_density)

        # f"plan_area_density test failed, expected {expected}, actual {actual}"
        pd.testing.assert_frame_equal(
            expected,
            actual,
        )

    def test_plan_area_fraction(self):
        """Test that the function `plan_area_fraction()` returns the correct value."""

        building_plan_area = pd.Series([0, 0.01, 70, 100])
        total_plan_area = pd.Series([1, 1, 0.1, 0])
        expected = pd.Series([0.0, 0.01, 700, math.inf])
        actual = nodes.plan_area_fraction(building_plan_area, total_plan_area)
        # f"plan_area_fraction test failed, expected {expected}, actual {actual}",
        pd.testing.assert_series_equal(
            expected,
            actual,
        )

    def test_raupach_displacement_height(self):
        """Test that the function `raupach_displacement_height()` returns the correct value for each cardinal direction."""

        constant_75 = Settings.CONSTANT_75
        building_height = pd.Series([1, 75])
        frontal_area_index = pd.DataFrame([[0.5, 0.0, 1, 10], [0.5, 0.0, 1, 10]])
        expected = pd.DataFrame(
            [
                [
                    1 - ((1 - math.exp(-math.sqrt(constant_75))) / math.sqrt(constant_75)),
                    math.nan,
                    1 - ((1 - math.exp(-math.sqrt(constant_75 * 2))) / math.sqrt(constant_75 * 2)),
                    1
                    - ((1 - math.exp(-math.sqrt(constant_75 * 20))) / math.sqrt(constant_75 * 20)),
                ],
                [
                    75 * (1 - ((1 - math.exp(-math.sqrt(constant_75))) / math.sqrt(constant_75))),
                    math.nan,
                    75
                    * (
                        1
                        - ((1 - math.exp(-math.sqrt(constant_75 * 2))) / math.sqrt(constant_75 * 2))
                    ),
                    75
                    * (
                        1
                        - (
                            (1 - math.exp(-math.sqrt(constant_75 * 20)))
                            / math.sqrt(constant_75 * 20)
                        )
                    ),
                ],
            ]
        )
        expected.columns = [
            Settings.RAUPACH_DISPLACEMENT_HEIGHT_NORTH,
            Settings.RAUPACH_DISPLACEMENT_HEIGHT_EAST,
            Settings.RAUPACH_DISPLACEMENT_HEIGHT_SOUTH,
            Settings.RAUPACH_DISPLACEMENT_HEIGHT_WEST,
        ]
        actual = nodes.raupach_displacement_height(building_height, frontal_area_index)
        # f"raupach_displacement_height test failed, expected {expected}, actual {actual}",
        pd.testing.assert_frame_equal(
            expected,
            actual,
        )

    def test_raupach_roughness_length(self):
        """Test that the function `raupach_roughness_length()` returns the correct values."""
        building_height = pd.Series([0, 0.5, 75])
        frontal_area_index = pd.DataFrame([[0, 0.5, 1.0, 75], [0, 0.5, 1.0, 75], [0, 0.5, 1.0, 75]])
        raupach_displacement_height = pd.DataFrame(
            [[0, 0.5, 1, 75], [0, 0.5, 1, 75], [0, 0.5, 1, 75]]
        )
        expected = pd.DataFrame(
            [
                [math.nan, math.nan, math.nan, math.nan],
                [0.0002776596113860747, 0.0, -0.1993248047654186, -56.45689024365426],
                [0.04164894170791121, 22.09119560552915, 29.500071105281958, 0.0],
            ]
        )
        expected.columns = [
            Settings.RAUPACH_ROUGHNESS_LENGTH_NORTH,
            Settings.RAUPACH_ROUGHNESS_LENGTH_EAST,
            Settings.RAUPACH_ROUGHNESS_LENGTH_SOUTH,
            Settings.RAUPACH_ROUGHNESS_LENGTH_WEST,
        ]
        actual = nodes.raupach_roughness_length(
            building_height, frontal_area_index, raupach_displacement_height
        )
        # f"raupach_roughness_length test failed, expected {expected}, actual {actual}",
        pd.testing.assert_frame_equal(
            expected,
            actual,
        )

    def test_rooftop_area_density(self):
        """Test that the function `rooftop_area_density()` returns the correct value."""

        plan_area_density = pd.DataFrame(
            [[0.0, 1, 0.1, math.nan, math.inf, 100.0, 3.14, 112358, 0.0, 75, 1, 1, 1, 1, 1]]
        )
        expected = pd.DataFrame(
            [
                [
                    0.0,
                    1.0,
                    0.1,
                    math.nan,
                    math.inf,
                    100.0,
                    3.14,
                    112358.0,
                    0.0,
                    75.0,
                    1.0,
                    1.0,
                    1.0,
                    1.0,
                    1.0,
                ]
            ]
        )
        columns_rooftop_area_density = [
            f"{Settings.ROOFTOP_AREA_DENSITY}_{i}"
            for i in range(int(Settings.MAX_BUILDING_HEIGHT / Settings.BUILDING_HEIGHT_INTERVAL))
        ]
        expected.columns = columns_rooftop_area_density
        actual = nodes.rooftop_area_density(plan_area_density)
        # f"rooftop_area_density test failed, expected {expected}, actual {actual}",
        pd.testing.assert_frame_equal(
            expected,
            actual,
        )

    def test_sky_view_factor(self):
        """Test that the function `sky_view_factor()` returns the correct value."""

        building_height = pd.Series([0, 0.1, 5.5, 5.5, 75])
        average_distance_between_buildings = pd.Series([1, 1, 1, 0, 1])
        expected = pd.Series(
            [
                1.0,
                0.9805806756909201,
                0.09053574604251861,
                6.123233995736766e-17,
                0.0066665185234566545,
            ]
        )
        actual = nodes.sky_view_factor(building_height, average_distance_between_buildings)
        # f"sky_view_factor test failed, expected {expected}, actual {actual}",
        pd.testing.assert_series_equal(
            expected,
            actual,
        )

    def test_standard_deviation_of_building_heights(self):
        """Test that the function `standard_deviation_of_building_heights()` returns the correct value."""

        buildings_intersecting_plan_area = gpd.GeoDataFrame(
            [[0, 10], [0, 1], [1, 0.1], [1, 0], [2, 112358], [0, 100]]
        )
        buildings_intersecting_plan_area.columns = [
            Settings.TARGET_ID_FIELD,
            Settings.NEIGHBOR_HEIGHT_FIELD,
        ]
        expected = pd.Series([54.74486277268398, 0.07071067811865477, 0])

        actual = nodes.standard_deviation_of_building_heights(buildings_intersecting_plan_area)
        # f"standard_deviation_of_building_heights test failed, expected {expected}, actual {actual}",
        pd.testing.assert_series_equal(
            expected,
            actual,
        )

    def test_wall_angle_direction_length(self):
        """Test that the function `wall_angle_direction_length()` returns the correct angle, direction, and length."""

        polygon_exterior = [[0, 1], [1, 1], [1, 0], [0, 0], [0, 1]]
        polygon_interior = [[0.25, 0.25], [0.25, 0.75], [0.75, 0.75], [0.75, 0.25]]

        north = Settings.NORTH
        south = Settings.SOUTH
        east = Settings.EAST
        west = Settings.WEST

        wall_angle = Settings.WALL_ANGLE
        wall_direction = Settings.WALL_DIRECTION
        wall_length = Settings.WALL_LENGTH

        square_root_one_half = 0.7071067811865476

        @dataclass
        class TestCase:
            name: str
            input: List[Polygon]
            expected: List[int]

        testcases = [
            TestCase(
                name="square",
                input=[Polygon(polygon_exterior)],
                expected=pd.concat(
                    [
                        pd.Series([[0.0, -90.0, 180.0, 90.0]], name=wall_angle),
                        pd.Series([[north, east, south, west]], name=wall_direction),
                        pd.Series([[1.0, 1.0, 1.0, 1.0]], name=wall_length),
                    ],
                    axis=1,
                ),
            ),
            TestCase(
                name="square with inner ring",
                input=[Polygon(polygon_exterior, [polygon_interior])],
                expected=pd.concat(
                    [
                        pd.Series([[0.0, -90.0, 180.0, 90.0]], name=wall_angle),
                        pd.Series([[north, east, south, west]], name=wall_direction),
                        pd.Series([[1.0, 1.0, 1.0, 1.0]], name=wall_length),
                    ],
                    axis=1,
                ),
            ),
            TestCase(
                name="45 degree triangle",
                input=[
                    Polygon(
                        [
                            [0, 0],
                            [square_root_one_half, square_root_one_half],
                            [0, square_root_one_half],
                        ]
                    )
                ],
                expected=pd.concat(
                    [
                        pd.Series([[45.0, 180.0, -90.0]], name=wall_angle),
                        pd.Series([[west, south, east]], name=wall_direction),
                        pd.Series(
                            [[1.0, square_root_one_half, square_root_one_half]], name=wall_length
                        ),
                    ],
                    axis=1,
                ),
            ),
            TestCase(
                name="135 degree triangle",
                input=[
                    Polygon(
                        [
                            [0, 0],
                            [-square_root_one_half, square_root_one_half],
                            [0, square_root_one_half],
                        ]
                    )
                ],
                expected=pd.concat(
                    [
                        pd.Series([[135.0, 0.0, -90.0]], name=wall_angle),
                        pd.Series([[south, north, east]], name=wall_direction),
                        pd.Series(
                            [[1.0, square_root_one_half, square_root_one_half]], name=wall_length
                        ),
                    ],
                    axis=1,
                ),
            ),
            TestCase(
                name="225 degree triangle",
                input=[
                    Polygon(
                        [
                            [0, 0],
                            [-square_root_one_half, -square_root_one_half],
                            [-square_root_one_half, 0],
                        ]
                    )
                ],
                expected=pd.concat(
                    [
                        pd.Series([[-135.0, 90.0, 0.0]], name=wall_angle),
                        pd.Series([[east, west, north]], name=wall_direction),
                        pd.Series(
                            [[1.0, square_root_one_half, square_root_one_half]], name=wall_length
                        ),
                    ],
                    axis=1,
                ),
            ),
            TestCase(
                name="325 degree triangle",
                input=[
                    Polygon(
                        [
                            [0, 0],
                            [square_root_one_half, -square_root_one_half],
                            [0, -square_root_one_half],
                        ]
                    )
                ],
                expected=pd.concat(
                    [
                        pd.Series([[-45.0, 180.0, 90.0]], name=wall_angle),
                        pd.Series([[north, south, west]], name=wall_direction),
                        pd.Series(
                            [[1.0, square_root_one_half, square_root_one_half]], name=wall_length
                        ),
                    ],
                    axis=1,
                ),
            ),
        ]

        for case in testcases:
            actual = nodes.wall_angle_direction_length(gpd.GeoSeries(case.input))
            expected = case.expected
            pd.testing.assert_frame_equal(
                expected,
                actual,
                f"wall_angle_direction_length test {case.name} failed, expected {expected}, actual {actual}",
            )

    def test_vertical_distribution_of_building_heights(self):
        """Test that the function `vertical_distribution_of_building_heights()` returns the correct dataframe."""
        building_height = pd.Series([0, 5, 5, 6, 7.5, 75])
        [
            f"{Settings.VERTICAL_DISTRIBUTION_OF_BUILDING_HEIGHTS}_{i}"
            for i in range(int(Settings.MAX_BUILDING_HEIGHT / Settings.BUILDING_HEIGHT_INTERVAL))
        ]
        expected = pd.DataFrame(
            [
                [0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0],
                [1.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0],
                [1.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0],
                [1.0, 1.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0],
                [1.0, 1.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0],
                [1.0, 1.0, 1.0, 1.0, 1.0, 1.0, 1.0, 1.0, 1.0, 1.0, 1.0, 1.0, 1.0, 1.0, 1.0],
            ],
            columns=[
                f"{Settings.VERTICAL_DISTRIBUTION_OF_BUILDING_HEIGHTS}_{i}"
                for i in range(
                    int(Settings.MAX_BUILDING_HEIGHT / Settings.BUILDING_HEIGHT_INTERVAL)
                )
            ],
        )
        actual = nodes.vertical_distribution_of_building_heights(building_height)
        # f"vertical_distribution_of_building_heights test failed, expected {expected}, actual {actual}",
        pd.testing.assert_frame_equal(
            expected,
            actual,
        )

    def test_wall_length(self):
        """Test that the function `wall_length()` returns the correct length."""

        north = Settings.NORTH
        south = Settings.SOUTH
        east = Settings.EAST
        west = Settings.WEST

        wall_angle = Settings.WALL_ANGLE
        wall_direction = Settings.WALL_DIRECTION
        wall_length = Settings.WALL_LENGTH

        wall_length_north = Settings.WALL_LENGTH_NORTH
        wall_length_south = Settings.WALL_LENGTH_SOUTH
        wall_length_east = Settings.WALL_LENGTH_EAST
        wall_length_west = Settings.WALL_LENGTH_WEST

        square_root_one_half = 0.7071067811865476

        square_input = pd.concat(
            [
                pd.Series([[0.0, -90.0, 180.0, 90.0]], name=wall_angle),
                pd.Series([[north, east, south, west]], name=wall_direction),
                pd.Series([[1.0, 1.0, 1.0, 1.0]], name=wall_length),
            ],
            axis=1,
        )
        triangle_input = pd.concat(
            [
                pd.Series([[45.0, 180.0, -90.0]], name=wall_angle),
                pd.Series([[west, south, east]], name=wall_direction),
                pd.Series([[1.0, square_root_one_half, square_root_one_half]], name=wall_length),
            ],
            axis=1,
        )
        eight_sided_input = pd.concat(
            [
                pd.Series([[0.0, -90.0, 180.0, 90.0, -45.0, -135.0, 135.0, 45.0]], name=wall_angle),
                pd.Series(
                    [[north, east, south, west, north, east, south, west]], name=wall_direction
                ),
                pd.Series([[1.0, 2.0, 3.0, 4.0, 5.0, 6.0, 7.0, 8.0]], name=wall_length),
            ],
            axis=1,
        )

        @dataclass
        class TestCase:
            name: str
            input: pd.DataFrame
            expected: pd.DataFrame

        testcases = [
            TestCase(
                name="square",
                input=square_input,
                expected=pd.concat(
                    [
                        pd.Series([1.0], name=wall_length_north),
                        pd.Series([1.0], name=wall_length_east),
                        pd.Series([1.0], name=wall_length_south),
                        pd.Series([1.0], name=wall_length_west),
                    ],
                    axis=1,
                ),
            ),
            TestCase(
                name="triangle",
                input=triangle_input,
                expected=pd.concat(
                    [
                        pd.Series([0], name=wall_length_north),
                        pd.Series([square_root_one_half], name=wall_length_east),
                        pd.Series([square_root_one_half], name=wall_length_south),
                        pd.Series([1.0], name=wall_length_west),
                    ],
                    axis=1,
                ),
            ),
            TestCase(
                name="eight sided",
                input=eight_sided_input,
                expected=pd.concat(
                    [
                        pd.Series([6.0], name=wall_length_north),
                        pd.Series([8.0], name=wall_length_east),
                        pd.Series([10.0], name=wall_length_south),
                        pd.Series([12.0], name=wall_length_west),
                    ],
                    axis=1,
                ),
            ),
        ]

        for case in testcases:
            actual = nodes.wall_length(case.input)
            expected = case.expected
            pd.testing.assert_frame_equal(
                expected,
                actual,
                f"wall_length test {case.name} failed, expected {expected}, actual {actual}",
            )


if __name__ == "__main__":
    unittest.main()



import geopandas as gpd
import math
import numpy as np
import pandas as pd
from pyproj.crs import CRS
from hamilton.function_modifiers import extract_columns

from .config import Settings


def area_weighted_mean_of_building_heights(
    buildings_intersecting_plan_area: gpd.GeoDataFrame,
) -> pd.Series:
    """Calculate the area weighted mean of building heights for each target building in a GeoPandas GeoDataFrame.
    The entire area of buildings considered to be neighbors are included in the calculation.

    :param buildings_intersecting_plan_area:    Geometry field for the neighboring buildings from the spatially
                                                joined data.
    :type buildings_intersecting_plan_area:     gpd.GeoDataFrame

    :return:                                    The area weighted mean of building heights for all buildings within the target
                                                building's plan area.
    """

    buildings_intersecting_plan_area[Settings.NEIGHBOR_VOLUME_FIELD] = (
        buildings_intersecting_plan_area[Settings.NEIGHBOR_HEIGHT_FIELD]
        * buildings_intersecting_plan_area[Settings.NEIGHBOR_AREA_FIELD]
    )

    volume_sum = buildings_intersecting_plan_area.groupby(Settings.TARGET_ID_FIELD)[
        Settings.NEIGHBOR_VOLUME_FIELD
    ].sum()
    area_sum = buildings_intersecting_plan_area.groupby(Settings.TARGET_ID_FIELD)[
        Settings.NEIGHBOR_AREA_FIELD
    ].sum()

    df = volume_sum / area_sum

    return pd.Series(df.values)


def average_distance_between_buildings(distance_between_buildings: pd.Series) -> pd.Series:
    """Calculate the average distance from the target building to all neighboring buildings.

    :param distance_between_buildings:          distance from the target building to each
                                                neighbor building.
    :type distance_between_buildings:           pd.Series

    :return:                                    float

    """

    df = distance_between_buildings.replace(0, np.nan)
    df = (
        df.groupby(Settings.NEIGHBOR_ID_FIELD)
        .mean()
        .reset_index()
        .replace(np.nan, Settings.DEFAULT_STREET_WIDTH)
        .rename(
            columns={
                Settings.DISTANCE_BETWEEN_BUILDINGS: Settings.AVERAGE_DISTANCE_BETWEEN_BUILDINGS
            }
        )
    )
    return df[Settings.AVERAGE_DISTANCE_BETWEEN_BUILDINGS]


def building_area(building_geometry: pd.Series) -> pd.Series:
    """Calculate the area of the building geometry.

    :param building_geometry:                       Building Geometry.
    :type building_geometry:                        pd.Series

    :return:                                        pd.Series

    """

    return building_geometry.area


def buildings_intersecting_plan_area(
    building_id: pd.Series,
    building_height: pd.Series,
    building_geometry: pd.Series,
    building_area: pd.Series,
    total_plan_area_geometry: pd.Series,
    wall_length: pd.DataFrame,
    target_crs: CRS,
    join_type: str = "left",
    join_predicate: str = "intersects",
    join_lsuffix: str = Settings.TARGET,
    join_rsuffix: str = Settings.NEIGHBOR,
) -> gpd.GeoDataFrame:
    """Conduct a spatial join to get the building areas that intersect the buffered target buildings.

    :param building_id:                         Building ID field.
    :type building_id:                          pd.Series

    :param building_height:                     Building height field.
    :type building_height:                      pd.Series

    :param building_geometry:                   Geometry field for the buildings.
    :type building_geometry:                    pd.Series

    :param building_area:                       Building area field.
    :type building_area:                        pd.Series

    :param total_plan_area_geometry:            Geometry of the buffered building.
    :type total_plan_area_geometry:             pd.Series

    :param target_crs:                          Coordinate reference system field of the parent geometry.
    :type target_crs:                           pd.Series

    :param join_type:                           Type of join desired.
                                                DEFAULT: `left`
    :type join_type:                            str

    :param join_predicate:                      Selected topology of join.
                                                DEFAULT: `intersects`
    :type join_predicate:                       str

    :param join_lsuffix:                        Suffix of the left object in the join.
                                                DEFAULT: `target`
    :type join_lsuffix:                         str

    :param join_rsuffix:                        Suffix of the right object in the join.
                                                DEFAULT: `neighbor`
    :type join_rsuffix:                         str

    :return:                                    GeoDataFrame of building areas that intersect the buffered target
                                                buildings and their attributes.

    """

    df = pd.DataFrame(
        {
            Settings.ID_FIELD: building_id,
            Settings.HEIGHT_FIELD: building_height,
            Settings.AREA_FIELD: building_area,
            Settings.GEOMETRY_FIELD: building_geometry,
            Settings.BUFFERED_FIELD: total_plan_area_geometry,
            Settings.WALL_LENGTH_NORTH: wall_length[Settings.WALL_LENGTH_NORTH],
            Settings.WALL_LENGTH_EAST: wall_length[Settings.WALL_LENGTH_EAST],
            Settings.WALL_LENGTH_SOUTH: wall_length[Settings.WALL_LENGTH_SOUTH],
            Settings.WALL_LENGTH_WEST: wall_length[Settings.WALL_LENGTH_WEST],
        }
    )

    # Create left and right GeoDataFrames.
    left_gdf = gpd.GeoDataFrame(df, geometry=Settings.BUFFERED_FIELD, crs=target_crs)
    right_gdf = gpd.GeoDataFrame(df, geometry=Settings.GEOMETRY_FIELD, crs=target_crs)

    # Spatially join the building areas to the target buffered areas.
    xdf = gpd.sjoin(
        left_df=left_gdf,
        right_df=right_gdf,
        how=join_type,
        predicate=join_predicate,
        lsuffix=join_lsuffix,
        rsuffix=join_rsuffix,
    )

    # Add the neighbor building geometry.
    xdf = (
        xdf.set_index(f"{Settings.ID_FIELD}_{join_rsuffix}")
        .join(
            right_gdf.set_index(Settings.ID_FIELD)[Settings.GEOMETRY_FIELD].rename(
                Settings.NEIGHBOR_GEOMETRY_FIELD
            )
        )
        .sort_index()
    )

    return gpd.GeoDataFrame(xdf).set_geometry(Settings.GEOMETRY_FIELD)


def building_plan_area(
    buildings_intersecting_plan_area: gpd.GeoDataFrame,
    join_predicate: str = "intersection",
    join_rsuffix: str = Settings.NEIGHBOR,
) -> pd.Series:
    """Calculate the building plan area from the GeoDataFrame of buildings intersecting the plan area.

    :param buildings_intersecting_plan_area:    Geometry field for the neighboring buildings from the spatially
                                                joined data.
    :type buildings_intersecting_plan_area:     gpd.GeoDataFrame

    :param join_predicate:                      Selected topology of join.
                                                DEFAULT: `intersection`
    :type join_predicate:                       str

    :param join_rsuffix:                        Suffix of the right object in the join.
                                                DEFAULT: `neighbor`
    :type join_rsuffix:                         str

    :return:                                    The building plan area for each unique building in the
                                                `buildings_intersecting_plan_area` GeoDataFrame.

    """

    building_plan_area = []
    index = 0

    for target_building_id in np.sort(buildings_intersecting_plan_area.building_id_target.unique()):
        # Get DataFrame with any building that intersects the target_building_id plan area.
        target_building_gdf = buildings_intersecting_plan_area.loc[
            buildings_intersecting_plan_area[Settings.TARGET_ID_FIELD] == target_building_id
        ].reset_index()

        # Create GeoDataFrames with building and neighbor info.
        target_gdf = (
            target_building_gdf[[Settings.TARGET_ID_FIELD, Settings.TARGET_BUFFERED_FIELD]]
            .set_geometry(Settings.TARGET_BUFFERED_FIELD)
            .drop_duplicates()
        )
        neighbor_gdf = target_building_gdf[
            [f"index_{join_rsuffix}", Settings.NEIGHBOR_GEOMETRY_FIELD]
        ].set_geometry(Settings.NEIGHBOR_GEOMETRY_FIELD)

        # Create a new GeoDataFrame with the area of intersection.
        intersection_gdf = gpd.overlay(
            target_gdf, neighbor_gdf, how=join_predicate, keep_geom_type=False
        )

        # Sum up the area of intersection and add to the output list.
        building_plan_area.append(intersection_gdf[Settings.DATA_GEOMETRY_FIELD_NAME].area.sum())

        index += 1

    return pd.Series(building_plan_area)


def building_surface_area(
    wall_length: pd.DataFrame, building_height: pd.Series, building_area: pd.Series
) -> pd.Series:
    """Calculate the building surface area for each building in a Panda Series. In naturf, the building footprint area is the
    same as the roof area.

    :param wall_length:                   Wall length in each cardinal direction for each building.
    :type wall_length:                    pd.DataFrame

    :param building_height:               Building height for each building.
    :type building_height:                pd.Series

    :param building_area:                 Building area field.
    :type building_area:                  pd.Series

    :return:                              Panda Series with building surface area.
    """

    wall_area = wall_length.mul(building_height, axis=0).sum(axis=1)

    return wall_area + building_area


def building_surface_area_to_plan_area_ratio(
    building_surface_area: pd.Series, total_plan_area: pd.Series
) -> pd.Series:
    """Calculate the building surface area to plan area ratio for each building in a Pandas Series. In naturf, the building footprint area is the
    same as the roof area.

    :param building_surface_area:         Building surface area for each building.
    :type building_surface_area:          pd.Series

    :param total_plan_area:               Total plan area for each building.
    :type total_plan_area:                pd.Series

    :return:                              Panda Series with building surface area to plan area ratio.
    """

    return building_surface_area / total_plan_area


def complete_aspect_ratio(
    building_surface_area: pd.Series, total_plan_area: pd.Series, building_plan_area: pd.Series
) -> pd.Series:
    """Calculate the complete aspect ratio for each building in a Pandas Series. In naturf, the building footprint area is the
    same as the roof area, and the exposed ground is the difference between total plan area and building plan area.

    :param building_surface_area:         Building surface area for each building.
    :type building_surface_area:          pd.Series

    :param total_plan_area:               Total plan area for each building.
    :type total_plan_area:                pd.Series

    :param building_plan_area:            Building plan area for each building.
    :type building_plan_area:             pd.Series

    :return:                              Panda Series with complete aspect ratio.
    """

    exposed_ground = total_plan_area - building_plan_area

    return (building_surface_area + exposed_ground) / total_plan_area


def distance_between_buildings(buildings_intersecting_plan_area: gpd.GeoDataFrame) -> pd.Series:
    """Calculate the distance between each building and its neighbor as defined in buildings_intersecting_plan_area.

    :param buildings_intersecting_plan_area:    Geometry field for the neighboring buildings from the spatially
                                                joined data.
    :type buildings_intersecting_plan_area:     gpd.GeoDataFrame

    :return:                                    The distance between each building and its neighbors in a Pandas Series.
    """

    neighbor_geometry = buildings_intersecting_plan_area[Settings.NEIGHBOR_GEOMETRY_FIELD]

    return buildings_intersecting_plan_area.distance(neighbor_geometry).rename(
        Settings.DISTANCE_BETWEEN_BUILDINGS
    )


@extract_columns(*[Settings.ID_FIELD, Settings.HEIGHT_FIELD, Settings.GEOMETRY_FIELD])
def filter_height_range(standardize_column_names_df: gpd.GeoDataFrame) -> gpd.GeoDataFrame:
    """Filter out any zero height buildings and reindex the data frame.  Extract the building_id,
    building_height, and geometry fields to nodes.

    :param standardize_column_names_df:             GeoDataFrame of the input shapefile with renamed columns.
    :type standardize_column_names_df:              gpd.GeoDataFrame

    :return:                                        GeoDataFrame

    """

    standardize_column_names_df.loc[
        standardize_column_names_df[Settings.HEIGHT_FIELD] > Settings.MAX_BUILDING_HEIGHT,
        Settings.HEIGHT_FIELD,
    ] = Settings.MAX_BUILDING_HEIGHT

    return standardize_column_names_df.loc[
        standardize_column_names_df[Settings.HEIGHT_FIELD] > 0
    ].reset_index(drop=True)


def frontal_area(frontal_length: pd.DataFrame, building_height: pd.Series) -> pd.DataFrame:
    """Calculate the frontal area for each building in a Pandas DataFrame in each cardinal direction.

    :param frontal_length:                Frontal length in each cardinal direction for each building.
    :type frontal_length:                 pd.DataFrame

    :param building_height:               Building height for each building.
    :type building_height:                pd.Series

    :return:                              Pandas DataFrame with frontal area in each cardinal direction.
    """

    frontal_area = frontal_length.mul(building_height, axis=0)

    frontal_area.columns = [
        Settings.FRONTAL_AREA_NORTH,
        Settings.FRONTAL_AREA_EAST,
        Settings.FRONTAL_AREA_SOUTH,
        Settings.FRONTAL_AREA_WEST,
    ]

    return frontal_area


def frontal_area_density(
    frontal_length: pd.DataFrame, building_height: pd.Series, total_plan_area: pd.Series
) -> pd.DataFrame:
    """Calculate the frontal area density for each building in a GeoPandas GeoSeries. Frontal area density is the frontal area of a
    building at a specific height increment divided by the total plan area. naturf calculates frontal area density from the four cardinal
    directions (east, north, west, south) and at 5 meter increments from ground level to 75 meters unless otherwise specified.

    :param frontal_length:                Frontal length in each cardinal direction for each building.
    :type frontal_length:                 pd.DataFrame

    :param building_height:               Building height for each building.
    :type building_height:                pd.Series

    :param total_plan_area:               Total plan area for each building.
    :type total_plan_area:                pd.Series

    :return:                              Pandas DataFrame with frontal area density for each cardinal direction and
                                          each BUILDING_HEIGHT_INTERVAL for each building.
    """

    rows, cols = (
        len(building_height.index),
        int(Settings.MAX_BUILDING_HEIGHT / Settings.BUILDING_HEIGHT_INTERVAL),
    )
    frontal_area_north, frontal_area_east, frontal_area_south, frontal_area_west = (
        [[0 for i in range(cols)] for j in range(rows)],
        [[0 for i in range(cols)] for j in range(rows)],
        [[0 for i in range(cols)] for j in range(rows)],
        [[0 for i in range(cols)] for j in range(rows)],
    )

    for building, building_frontal_length in frontal_length.iterrows():
        building_height_counter = Settings.BUILDING_HEIGHT_INTERVAL

        # Go from ground level to building height by the building height interval and calculate frontal area density.
        for building_height_index in range(
            0, math.ceil(building_height[building]), Settings.BUILDING_HEIGHT_INTERVAL
        ):
            if building_height_counter <= building_height[building]:
                frontal_area_north[building][
                    int(building_height_index / Settings.BUILDING_HEIGHT_INTERVAL)
                ] = (
                    building_frontal_length[Settings.FRONTAL_LENGTH_NORTH]
                    * Settings.BUILDING_HEIGHT_INTERVAL
                    / total_plan_area[building]
                )
                frontal_area_east[building][
                    int(building_height_index / Settings.BUILDING_HEIGHT_INTERVAL)
                ] = (
                    building_frontal_length[Settings.FRONTAL_LENGTH_EAST]
                    * Settings.BUILDING_HEIGHT_INTERVAL
                    / total_plan_area[building]
                )
                frontal_area_south[building][
                    int(building_height_index / Settings.BUILDING_HEIGHT_INTERVAL)
                ] = (
                    building_frontal_length[Settings.FRONTAL_LENGTH_SOUTH]
                    * Settings.BUILDING_HEIGHT_INTERVAL
                    / total_plan_area[building]
                )
                frontal_area_west[building][
                    int(building_height_index / Settings.BUILDING_HEIGHT_INTERVAL)
                ] = (
                    building_frontal_length[Settings.FRONTAL_LENGTH_WEST]
                    * Settings.BUILDING_HEIGHT_INTERVAL
                    / total_plan_area[building]
                )
            else:
                frontal_area_north[building][
                    int(building_height_index / Settings.BUILDING_HEIGHT_INTERVAL)
                ] = (
                    building_frontal_length[Settings.FRONTAL_LENGTH_NORTH]
                    * (
                        Settings.BUILDING_HEIGHT_INTERVAL
                        - building_height_counter
                        + building_height[building]
                    )
                    / total_plan_area[building]
                )
                frontal_area_east[building][
                    int(building_height_index / Settings.BUILDING_HEIGHT_INTERVAL)
                ] = (
                    building_frontal_length[Settings.FRONTAL_LENGTH_EAST]
                    * (
                        Settings.BUILDING_HEIGHT_INTERVAL
                        - building_height_counter
                        + building_height[building]
                    )
                    / total_plan_area[building]
                )
                frontal_area_south[building][
                    int(building_height_index / Settings.BUILDING_HEIGHT_INTERVAL)
                ] = (
                    building_frontal_length[Settings.FRONTAL_LENGTH_SOUTH]
                    * (
                        Settings.BUILDING_HEIGHT_INTERVAL
                        - building_height_counter
                        + building_height[building]
                    )
                    / total_plan_area[building]
                )
                frontal_area_west[building][
                    int(building_height_index / Settings.BUILDING_HEIGHT_INTERVAL)
                ] = (
                    building_frontal_length[Settings.FRONTAL_LENGTH_WEST]
                    * (
                        Settings.BUILDING_HEIGHT_INTERVAL
                        - building_height_counter
                        + building_height[building]
                    )
                    / total_plan_area[building]
                )
                break
            building_height_counter += Settings.BUILDING_HEIGHT_INTERVAL

    columns_north, columns_east, columns_south, columns_west = (
        [
            f"{Settings.FRONTAL_AREA_NORTH}_{i}"
            for i in range(int(Settings.MAX_BUILDING_HEIGHT / Settings.BUILDING_HEIGHT_INTERVAL))
        ],
        [
            f"{Settings.FRONTAL_AREA_EAST}_{i}"
            for i in range(int(Settings.MAX_BUILDING_HEIGHT / Settings.BUILDING_HEIGHT_INTERVAL))
        ],
        [
            f"{Settings.FRONTAL_AREA_SOUTH}_{i}"
            for i in range(int(Settings.MAX_BUILDING_HEIGHT / Settings.BUILDING_HEIGHT_INTERVAL))
        ],
        [
            f"{Settings.FRONTAL_AREA_WEST}_{i}"
            for i in range(int(Settings.MAX_BUILDING_HEIGHT / Settings.BUILDING_HEIGHT_INTERVAL))
        ],
    )

    return pd.concat(
        [
            pd.DataFrame(frontal_area_north, columns=columns_north),
            pd.DataFrame(frontal_area_east, columns=columns_east),
            pd.DataFrame(frontal_area_south, columns=columns_south),
            pd.DataFrame(frontal_area_west, columns=columns_west),
        ],
        axis=1,
    )


def frontal_area_index(frontal_area: pd.DataFrame, total_plan_area: pd.Series) -> pd.DataFrame:
    """Calculate the frontal area index for each building in a Pandas DataFrame in each cardinal direction.

    :param frontal_area:                  Frontal area in each cardinal direction for each building.
    :type frontal_area:                   pd.DataFrame

    :param total_plan_area:               Total plan area for each building.
    :type total_plan_area:                pd.Series

    :return:                              Pandas DataFrame with frontal area index in each cardinal direction.
    """

    frontal_area_index = frontal_area.div(total_plan_area, axis=0)
    frontal_area_index.columns = [
        Settings.FRONTAL_AREA_INDEX_NORTH,
        Settings.FRONTAL_AREA_INDEX_EAST,
        Settings.FRONTAL_AREA_INDEX_SOUTH,
        Settings.FRONTAL_AREA_INDEX_WEST,
    ]

    return frontal_area_index


def frontal_length(
    buildings_intersecting_plan_area: gpd.GeoDataFrame,
) -> pd.DataFrame:
    """Calculate the frontal length for each cardinal direction from the GeoDataFrame of buildings intersecting the plan area.
    `buildings_intersecting_plan_area()` needs to include `wall_length`.

    :param buildings_intersecting_plan_area:    Geometry field for the neighboring buildings from the spatially
                                                joined data.
    :type buildings_intersecting_plan_area:     gpd.GeoDataFrame

    :return:                                    The frontal area for each cardinal direction for each unique building in the
                                                `buildings_intersecting_plan_area` GeoDataFrame.

    """

    number_target_buildings = len(buildings_intersecting_plan_area.building_id_target.unique())
    frontal_length_north, frontal_length_east, frontal_length_south, frontal_length_west = (
        [0] * number_target_buildings,
        [0] * number_target_buildings,
        [0] * number_target_buildings,
        [0] * number_target_buildings,
    )
    index = 0

    for target_building_id in np.sort(buildings_intersecting_plan_area.building_id_target.unique()):
        # Get DataFrame with any building that intersects the target_building_id plan area.
        target_building_gdf = buildings_intersecting_plan_area.loc[
            buildings_intersecting_plan_area[Settings.TARGET_ID_FIELD] == target_building_id
        ].reset_index()

        # Sum frontal length for each cardinal direction
        frontal_length_north[index] = target_building_gdf[
            f"{Settings.WALL_LENGTH_NORTH}_{Settings.NEIGHBOR}"
        ].sum()
        frontal_length_east[index] = target_building_gdf[
            f"{Settings.WALL_LENGTH_EAST}_{Settings.NEIGHBOR}"
        ].sum()
        frontal_length_south[index] = target_building_gdf[
            f"{Settings.WALL_LENGTH_SOUTH}_{Settings.NEIGHBOR}"
        ].sum()
        frontal_length_west[index] = target_building_gdf[
            f"{Settings.WALL_LENGTH_WEST}_{Settings.NEIGHBOR}"
        ].sum()

        index += 1

    return pd.concat(
        [
            pd.Series(frontal_length_north, name=Settings.FRONTAL_LENGTH_NORTH),
            pd.Series(frontal_length_east, name=Settings.FRONTAL_LENGTH_EAST),
            pd.Series(frontal_length_south, name=Settings.FRONTAL_LENGTH_SOUTH),
            pd.Series(frontal_length_west, name=Settings.FRONTAL_LENGTH_WEST),
        ],
        axis=1,
    )


def grimmond_oke_displacement_height(building_height: pd.Series) -> pd.Series:
    """Calculate the Grimmond & Oke displacement height for each building

    :param building_height:                     Building height field.
    :type building_height:                      pd.Series

    :return:                                    pd.Series

    """

    return building_height * Settings.DISPLACEMENT_HEIGHT_FACTOR


def grimmond_oke_roughness_length(building_height: pd.Series) -> pd.Series:
    """Calculate the Grimmond & Oke roughness length for each building

    :param building_height:                     Building height field.
    :type building_height:                      pd.Series

    :return:                                    pd.Series

    """

    return building_height * Settings.ROUGHNESS_LENGTH_FACTOR


def height_to_width_ratio(
    mean_building_height: pd.Series, average_distance_between_buildings: pd.Series
) -> pd.Series:
    """Calculate the height to width ratio for each building.

    :param mean_building_height:           Series of mean building height for all buildings within the target building's plan area
    :type mean_building_height:            pd.Series

    :param average_distance_between_buildings: Series of average distance from each building to all neighboring buildings
    :type average_distance_between_buildings:  pd.Series

    :return:                                   pd.Series

    """

    return mean_building_height / average_distance_between_buildings


def input_shapefile_df(input_shapefile: str) -> gpd.GeoDataFrame:
    """Import shapefile to GeoDataFrame using only desired columns.

    :params input_shapefile:                        Full path with file name and extension to the input shapefile.
    :type input_shapefile:                          str

    :return:                                        GeoDataFrame

    """

    gdf = gpd.read_file(input_shapefile)[
        [
            Settings.DATA_ID_FIELD_NAME,
            Settings.DATA_HEIGHT_FIELD_NAME,
            Settings.DATA_GEOMETRY_FIELD_NAME,
        ]
    ].set_geometry(Settings.DATA_GEOMETRY_FIELD_NAME)

    return gdf


def lot_area(
    buildings_intersecting_plan_area: gpd.GeoDataFrame, building_surface_area: pd.Series
) -> pd.Series:
    """Calculate the lot area for each building in a Panda Series. Lot area is the total surface area of all buildings
    within a given building's plan area divided by the number of buildings in the plan area."

    :param buildings_intersecting_plan_area:    Geometry field for the neighboring buildings from the spatially
                                                joined data.
    :type buildings_intersecting_plan_area:     gpd.GeoDataFrame

    :param building_surface_area:               Building surface area for each building.
    :type building_surface_area:                pd.Series

    :return:                                    Panda Series of lot area for each building.
    """

    df = buildings_intersecting_plan_area.join(
        building_surface_area.rename(Settings.BUILDING_SURFACE_AREA),
        on=f"index_{Settings.NEIGHBOR}",
        how="left",
    )
    df = df.groupby(Settings.TARGET_ID_FIELD)[Settings.BUILDING_SURFACE_AREA].mean()

    return pd.Series(df.values)


def macdonald_displacement_height(
    building_height: pd.Series, plan_area_fraction: pd.Series
) -> pd.Series:
    """Calculate the Macdonald et al. displacement height for each building in a Pandas Series.

    :param building_height:               Building height for each building.
    :type building_height:                pd.Series

    :param plan_area_fraction:            Plan area fraction for each building.
    :type plan_area_fraction:             pd.Series
    """

    alpha_coefficient = Settings.ALPHACOEFFICIENT

    plan_area_minus_one = plan_area_fraction - 1

    alpha_exponent = alpha_coefficient ** (-plan_area_fraction)

    right_side = 1 + alpha_exponent * plan_area_minus_one

    return right_side * building_height


def macdonald_roughness_length(
    building_height: pd.Series,
    macdonald_displacement_height: pd.Series,
    frontal_area: pd.DataFrame,
    lot_area: pd.Series,
) -> pd.DataFrame:
    """Calculate the Macdonald et al. roughness length for each building in a Pandas Series.

    :param building_height:               Building height for each building.
    :type building_height:                pd.Series

    :param macdonald_displacement_height: Macdonald displacement height for each building.
    :type macdonald_displacement_height:  pd.Series

    :param frontal_area:                  Frontal area  for each building in each cardinal direction.
    :type frontal_area:                   pd.DataFrame

    :param lot_area:                      Lot area for each building.
    :type lot_area:                       pd.Series

    :return:                              Panda Series with Macdonald roughness length for each building in each cardinal direction.
    """

    obstacle_drag_coefficient = Settings.OBSTACLEDRAGCOEFFICIENT
    von_karman_constant = Settings.VONKARMANCONSTANT
    beta_coefficient = Settings.BETACOEFFICIENT

    one_minus_height_ratio = 1 - macdonald_displacement_height / building_height

    drag_over_von_karman = obstacle_drag_coefficient / von_karman_constant**2

    frontal_divided_by_lot = frontal_area.div(lot_area, axis=0)

    inside_exponential = -frontal_divided_by_lot.mul(
        0.5 * beta_coefficient * drag_over_von_karman * one_minus_height_ratio, axis=0
    ) ** (-0.5)

    exponential = np.exp(inside_exponential)

    right_side = exponential.mul(one_minus_height_ratio, axis=0)

    macdonald_roughness_length = right_side.mul(building_height, axis=0)

    macdonald_roughness_length.columns = [
        Settings.MACDONALD_ROUGHNESS_LENGTH_NORTH,
        Settings.MACDONALD_ROUGHNESS_LENGTH_EAST,
        Settings.MACDONALD_ROUGHNESS_LENGTH_SOUTH,
        Settings.MACDONALD_ROUGHNESS_LENGTH_WEST,
    ]

    return macdonald_roughness_length


def mean_building_height(buildings_intersecting_plan_area: gpd.GeoDataFrame) -> pd.Series:
    """Calculate the mean building height for all buildings within the target building's total plan area.

    :param buildings_intersecting_plan_area:    Geometry field for the neighboring buildings from the spatially
                                                joined data.
    :type buildings_intersecting_plan_area:     gpd.GeoDataFrame

    :return:                                    The mean building height for all buildings within the target building's plan area.
    """

    df = buildings_intersecting_plan_area.groupby(Settings.TARGET_ID_FIELD)[
        Settings.NEIGHBOR_HEIGHT_FIELD
    ].mean()

    return pd.Series(df.values)


def plan_area_density(
    building_plan_area: pd.Series, building_height: pd.Series, total_plan_area: pd.Series
) -> pd.DataFrame:
    """Calculate the plan area density for each building in a GeoPandas GeoSeries. Plan area density is the building plan area
    at a specific height increment divided by the total plan area. naturf calculates plan area density from the four cardinal
    directions (north, east, south, west) and at 5-meter increments from ground level to 75 meters unless otherwise specified.

    :param building_plan_area:            Building plan area for each building.
    :type building_plan_area:             pd.Series

    :param building_height:               Building height for each building.
    :type building_height:                pd.Series

    :param total_plan_area:               Total plan area for each building.
    :type total_plan_area:                pd.Series

    :return:                              Pandas DataFrame with plan area density for each BUILDING_HEIGHT_INTERVAL for each building.
    """

    rows, cols = (
        len(building_height.index),
        int(Settings.MAX_BUILDING_HEIGHT / Settings.BUILDING_HEIGHT_INTERVAL),
    )
    plan_area_density = [[0 for i in range(cols)] for j in range(rows)]

    for building in range(building_plan_area.size):
        building_height_counter = 0

        # Go from ground level to building height by the building height interval and calculate plan area density.
        while building_height_counter < building_height[building]:
            plan_area_density[building][
                int(building_height_counter / Settings.BUILDING_HEIGHT_INTERVAL)
            ] = (building_plan_area[building] / total_plan_area[building])
            building_height_counter += Settings.BUILDING_HEIGHT_INTERVAL

    columns_plan_area_density = [
        f"{Settings.PLAN_AREA_DENSITY}_{i}"
        for i in range(int(Settings.MAX_BUILDING_HEIGHT / Settings.BUILDING_HEIGHT_INTERVAL))
    ]
    return pd.DataFrame(plan_area_density, columns=columns_plan_area_density)


def plan_area_fraction(building_plan_area: pd.Series, total_plan_area: pd.Series) -> pd.Series:
    """Calculate the plan area fraction for each building in a Pandas Series. Plan area fraction is the building plan area at ground level
    for each building divided by the total plan area.

    :param building_plan_area:            Building plan area for each building.
    :type building_plan_area:             pd.Series

    :param total_plan_area:               Total plan area for each building.
    :type total_plan_area:                pd.Series

    :return:                              Pandas Series with plan area fraction for each building.
    """

    return building_plan_area / total_plan_area


def raupach_displacement_height(
    building_height: pd.Series, frontal_area_index: pd.DataFrame
) -> pd.DataFrame:
    """Calculate the Raupach displacement height for each building in each cardinal direction in a Panda Series. Default values for constants are set
    in the config file.

    :param building_height:               Building height for each building.
    :type building_height:                pd.Series

    :param frontal_area_index:            Frontal area index for each building in each cardinal direction.
    :type frontal_area_index:             pd.DataFrame

    :return:                              Pandas DataFrame with Raupach displacement height in each cardinal direction.
    """

    constant_75 = Settings.CONSTANT_75
    capital_lambda = frontal_area_index.mul(2, axis=0)

    sqrt_constant_lambda = np.sqrt(capital_lambda.mul(constant_75, axis=0))
    numerator = 1 - np.exp(-sqrt_constant_lambda)

    raupach_displacement_height = (1 - numerator / sqrt_constant_lambda).mul(
        building_height, axis=0
    )

    raupach_displacement_height.columns = [
        Settings.RAUPACH_DISPLACEMENT_HEIGHT_NORTH,
        Settings.RAUPACH_DISPLACEMENT_HEIGHT_EAST,
        Settings.RAUPACH_DISPLACEMENT_HEIGHT_SOUTH,
        Settings.RAUPACH_DISPLACEMENT_HEIGHT_WEST,
    ]

    return raupach_displacement_height


def raupach_roughness_length(
    building_height: pd.Series,
    frontal_area_index: pd.DataFrame,
    raupach_displacement_height: pd.DataFrame,
) -> pd.DataFrame:
    """Calculate the Raupach roughness length for each building in each cardinal direction in a Panda Series. Default values for constants are set
    in the config file.

    :param building_height:               Building height for each building.
    :type building_height:                pd.Series

    :param frontal_area_index:            Frontal area index for each building in each cardinal direction.
    :type frontal_area_index:             pd.DataFrame

    :param raupach_displacment_height:    Raupach displacment height for each building in each cardinal direction.
    :type raupach_displacment_height:     pd.DataFrame

    :return:                              Pandas DataFrame with Raupach roughness length in each cardinal direction.
    """

    cols = [
        Settings.RAUPACH_ROUGHNESS_LENGTH_NORTH,
        Settings.RAUPACH_ROUGHNESS_LENGTH_EAST,
        Settings.RAUPACH_ROUGHNESS_LENGTH_SOUTH,
        Settings.RAUPACH_ROUGHNESS_LENGTH_WEST,
    ]

    cols_fai = frontal_area_index.columns
    cols_rdh = raupach_displacement_height.columns

    von_karman_constant = Settings.VONKARMANCONSTANT
    drag_coefficient_03 = Settings.DRAGCOEFFICIENT_03
    drag_coefficient_0003 = Settings.DRAGCOEFFICIENT_0003
    psi_h = Settings.PSI_H

    frontal_area_index.columns = cols
    raupach_displacement_height.columns = cols

    denominator = np.sqrt(drag_coefficient_0003 + drag_coefficient_03 * frontal_area_index)

    one_minus_displacement_height = 1 - raupach_displacement_height.div(building_height, axis=0)

    exponential = np.exp(-von_karman_constant / denominator - psi_h)

    raupach_roughness_length = exponential.mul(one_minus_displacement_height, axis=0).mul(
        building_height, axis=0
    )

    frontal_area_index.columns = cols_fai
    raupach_displacement_height.columns = cols_rdh

    return raupach_roughness_length


def rooftop_area_density(plan_area_density: pd.DataFrame) -> pd.DataFrame:
    """Calculate the rooftop area density for each building in a Pandas DataFrame. Rooftop area density is the roof area
    of all buildings within the total plan area  at a specified height increment divided by the total plan area. naturf
    projects building footprints vertically to the building height, meaning that rooftop area density is equal to the plan area
    density.

    :param plan_area_density:            Plan area density at each specified height increment.
    :type plan_area_density:             pd.DataFrame

    :return:                             Pandas DataFrame with rooftop area density for each BUILDING_HEIGHT_INTERVAL for each building.
    """

    columns_rooftop_area_density = [
        f"{Settings.ROOFTOP_AREA_DENSITY}_{i}"
        for i in range(int(Settings.MAX_BUILDING_HEIGHT / Settings.BUILDING_HEIGHT_INTERVAL))
    ]

    return pd.DataFrame(plan_area_density.values.tolist(), columns=columns_rooftop_area_density)


def sky_view_factor(
    building_height: pd.Series, average_distance_between_buildings: pd.Series
) -> pd.Series:
    """Calculate the 2D sky view factor for each building.

    :param building_height:                     Building height field.
    :type building_height:                      pd.Series

    :param average_distance_between_buildings:  Average distance between the target building and all neighboring buildings.
    :type average_distance_between_buildings:   float

    :return:                                    pd.Series

    """

    return np.cos(np.arctan(building_height / (0.5 * average_distance_between_buildings)))


def standard_deviation_of_building_heights(
    buildings_intersecting_plan_area: gpd.GeoDataFrame,
) -> pd.Series:
    """Calculate the standard deviation of building heights for all buildings within the target building's total plan area.

    :param buildings_intersecting_plan_area:    Geometry field for the neighboring buildings from the spatially
                                                joined data.
    :type buildings_intersecting_plan_area:     gpd.GeoDataFrame

    :return:                                    The standard deviation of building heights for all buildings within the target building's plan area.
    """

    df = (
        buildings_intersecting_plan_area.groupby(Settings.TARGET_ID_FIELD)[
            Settings.NEIGHBOR_HEIGHT_FIELD
        ]
        .std()
        .fillna(0)
    )

    return pd.Series(df.values)


def standardize_column_names_df(input_shapefile_df: gpd.GeoDataFrame) -> gpd.GeoDataFrame:
    """Standardize field names so use throughout code will be consistent throughout.

    :params input_shapefile_df:                     GeoDataFrame of from the input shapefile.
    :type input_shapefile_df:                       gpd.GeoDataFrame

    :return:                                        GeoDataFrame

    """

    # standardize field names from data to reference names in code
    input_shapefile_df.rename(
        columns={
            Settings.DATA_ID_FIELD_NAME: Settings.ID_FIELD,
            Settings.DATA_HEIGHT_FIELD_NAME: Settings.HEIGHT_FIELD,
            Settings.DATA_GEOMETRY_FIELD_NAME: Settings.GEOMETRY_FIELD,
        },
        inplace=True,
    )

    return input_shapefile_df.set_geometry(Settings.GEOMETRY_FIELD)


def target_crs(input_shapefile_df: gpd.GeoDataFrame) -> CRS:
    """Extract coordinate reference system from geometry.

    :params input_shapefile_df:                     GeoDataFrame of from the input shapefile.
    :type input_shapefile_df:                       gpd.GeoDataFrame

    :return:                                        pyproj Coordinate Reference System (CRS) object

    """

    return input_shapefile_df.set_geometry(Settings.GEOMETRY_FIELD).crs


def total_plan_area(total_plan_area_geometry: gpd.GeoSeries) -> pd.Series:
    """Calculate the total plan area for each building in a GeoPandas GeoSeries.

    :param geometry:                    Geometry for a series of buildings.
    :type geometry:                     gpd.GeoSeries

    :return:                            Pandas Series with total plan area for each building.

    """

    return total_plan_area_geometry.area


def total_plan_area_geometry(
    building_geometry: pd.Series, radius: int = Settings.RADIUS, cap_style: int = Settings.CAP_STYLE
) -> gpd.GeoSeries:
    """Calculate the geometry of the total plan area which is the buffer of the building for the desired radius and cap style.

    :param building_geometry:                       Geometry of the building.
    :type building_geometry:                        pd.Series

    :param radius:                                  The radius of the buffer.
                                                    100 (default, set in config.py)
    :type radius:                                   int

    :param cap_style:                               The shape of the buffer.
                                                    1 == Round
                                                    2 == Flat
                                                    3 == Square (default)

    :return:                                        pd.Series

    """

    return building_geometry.buffer(distance=radius, cap_style=cap_style)


def vertical_distribution_of_building_heights(building_height: pd.Series) -> pd.DataFrame:
    """Represent the location of buildings at 5m increments from ground level to 75m unless otherwise specified. If is within a
    given height bin, it will be given a 1 and it will be given a 0 otherwise."

    :param building_height:               Building height for each building.
    :type building_height:                pd.Series

    :return:                              Pandas DataFrame with the distribution of building heights at each
                                          BUILDING_HEIGHT_INTERVAL for each building.
    """

    rows, cols = (
        len(building_height.index),
        int(Settings.MAX_BUILDING_HEIGHT / Settings.BUILDING_HEIGHT_INTERVAL),
    )
    vertical_distribution_of_building_heights = [[0 for i in range(cols)] for j in range(rows)]

    for building in range(building_height.size):
        building_height_counter = 0

        while building_height_counter < building_height[building]:
            vertical_distribution_of_building_heights[building][
                int(building_height_counter / Settings.BUILDING_HEIGHT_INTERVAL)
            ] = 1.0
            building_height_counter += Settings.BUILDING_HEIGHT_INTERVAL

    columns_vertical_distribution_of_building_heights = [
        f"{Settings.VERTICAL_DISTRIBUTION_OF_BUILDING_HEIGHTS}_{i}"
        for i in range(int(Settings.MAX_BUILDING_HEIGHT / Settings.BUILDING_HEIGHT_INTERVAL))
    ]

    return pd.DataFrame(
        vertical_distribution_of_building_heights,
        columns=columns_vertical_distribution_of_building_heights,
    )


def wall_angle_direction_length(building_geometry: pd.Series) -> pd.DataFrame:
    """Calculate the wall angle, direction, and length for each building in a GeoPandas GeoSeries.

    :param geometry:                    Geometry for a series of buildings.
    :type geometry:                     pd.Series

    :return:                            Pandas DataFrame with wall angle, direction, and length for each building.

    """

    wall_angle, wall_direction, wall_length = (
        [[] for x in range(building_geometry.size)],
        [[] for x in range(building_geometry.size)],
        [[] for x in range(building_geometry.size)],
    )

    for building in range(building_geometry.size):
        points_in_polygon = building_geometry.values[building].exterior.xy

        for index, item in enumerate(zip(points_in_polygon[0], points_in_polygon[1])):
            x, y = item

            # Store the first set of coordinates.
            if index == 0:
                x1, y1 = x, y

            else:
                x2, y2 = x, y

                wall_angle[building].append(np.degrees(np.arctan2(y2 - y1, x2 - x1)))

                # For each direction, the start degree (from counterclockwise) is included (<=) and the end degree is not included (<).
                if (
                    Settings.NORTHEAST_DEGREES
                    <= wall_angle[building][index - 1]
                    < Settings.NORTHWEST_DEGREES
                ):
                    wall_direction[building].append(Settings.WEST)
                elif (
                    Settings.SOUTHEAST_DEGREES_ARCTAN
                    <= wall_angle[building][index - 1]
                    < Settings.NORTHEAST_DEGREES
                ):
                    wall_direction[building].append(Settings.NORTH)
                elif (
                    Settings.SOUTHWEST_DEGREES_ARCTAN
                    <= wall_angle[building][index - 1]
                    < Settings.SOUTHEAST_DEGREES_ARCTAN
                ):
                    wall_direction[building].append(Settings.EAST)
                else:
                    wall_direction[building].append(Settings.SOUTH)

                wall_length[building].append(np.sqrt((x2 - x1) ** 2 + (y2 - y1) ** 2))

                # Reset start coordinates.
                x1, y1 = x, y

    return pd.concat(
        [
            pd.Series(wall_angle, name=Settings.WALL_ANGLE),
            pd.Series(wall_direction, name=Settings.WALL_DIRECTION),
            pd.Series(wall_length, name=Settings.WALL_LENGTH),
        ],
        axis=1,
    )


def wall_length(wall_angle_direction_length: pd.DataFrame) -> pd.DataFrame:
    """Calculate the wall length for each building in a GeoPandas GeoSeries.

    :param wall_angle_direction_length:                Wall angle, direction, and length for a series of buildings.
    :type wall_angle_direction_length:                 pd.DataFrame

    :return:                                           Pandas DataFrame with wall area for each cardinal direction for each building.

    """

    wall_length_north, wall_length_east, wall_length_south, wall_length_west = (
        [0] * len(wall_angle_direction_length.index),
        [0] * len(wall_angle_direction_length.index),
        [0] * len(wall_angle_direction_length.index),
        [0] * len(wall_angle_direction_length.index),
    )

    for index, row in wall_angle_direction_length.iterrows():
        for j in range(len(row[Settings.WALL_DIRECTION])):
            if row[Settings.WALL_DIRECTION][j] == Settings.NORTH:
                wall_length_north[index] += row[Settings.WALL_LENGTH][j]
            elif row[Settings.WALL_DIRECTION][j] == Settings.EAST:
                wall_length_east[index] += row[Settings.WALL_LENGTH][j]
            elif row[Settings.WALL_DIRECTION][j] == Settings.SOUTH:
                wall_length_south[index] += row[Settings.WALL_LENGTH][j]
            else:
                wall_length_west[index] += row[Settings.WALL_LENGTH][j]

    return pd.concat(
        [
            pd.Series(wall_length_north, name=Settings.WALL_LENGTH_NORTH),
            pd.Series(wall_length_east, name=Settings.WALL_LENGTH_EAST),
            pd.Series(wall_length_south, name=Settings.WALL_LENGTH_SOUTH),
            pd.Series(wall_length_west, name=Settings.WALL_LENGTH_WEST),
        ],
        axis=1,
    )



import math
import os
import shutil
import tempfile
import unittest

import geopandas as gpd
import numpy as np
import pandas as pd
from shapely.geometry import Point, Polygon
import xarray as xr

import naturf.output as output
from naturf.config import Settings


class TestNodes(unittest.TestCase):
    def setUp(self):
        # Create a temporary directory
        self.test_dir = tempfile.mkdtemp()

    def tearDown(self):
        # Remove the directory after the test
        shutil.rmtree(self.test_dir)

    INPUTS = {
        "input_shapefile": os.path.join("naturf", "data", "C-5.shp"),
        "radius": 100,
        "cap_style": 1,
    }

    def test_aggregate_rasters(self):
        "Test that the function `aggregate_rasters()` returns the correct Xarray."
        rasterize_parameters = xr.Dataset(
            data_vars=dict(
                var0=(["x", "y"], np.array([[0, 1], [2, 3]])),
                var1=(["x", "y"], np.array([[10, 15.5], [314159, 75]])),
                building_count=(["x", "y"], np.array([[0, 1], [10, 75]])),
            ),
            coords=dict(
                lat=(["x"], [0, 1]),
                lon=(["y"], [0, 1]),
            ),
        )
        expected = xr.Dataset(
            data_vars=dict(
                var0=(["x", "y"], np.array([[0.0, 1.0], [0.2, 0.04]])),
                var1=(["x", "y"], np.array([[math.inf, 15.5], [31415.9, 1.0]])),
                building_count=(["x", "y"], np.array([[0.0, 1.0], [1.0, 1.0]])),
            ),
            coords=dict(
                lat=(["x"], [0, 1]),
                lon=(["y"], [0, 1]),
            ),
        )
        actual = output.aggregate_rasters(rasterize_parameters)
        xr.testing.assert_equal(expected, actual)

    def test_merge_parameters(self):
        """Test the function `merge_parameters()` to ensure it outputs the right type and shape GeoDataFrame."""

        frontal_area_density = pd.DataFrame(
            {
                Settings.FRONTAL_AREA_NORTH: [0.1, 0.2],
                Settings.FRONTAL_AREA_EAST: [0.3, 0.4],
                Settings.FRONTAL_AREA_SOUTH: [0.5, 0.6],
                Settings.FRONTAL_AREA_WEST: [0.7, 0.8],
            }
        )
        plan_area_density = pd.DataFrame({Settings.PLAN_AREA_DENSITY: [0.5, 0.6]})
        rooftop_area_density = pd.DataFrame({Settings.ROOFTOP_AREA_DENSITY: [0.9, 1.0]})
        plan_area_fraction = pd.Series([0.5, 0.5])
        mean_building_height = pd.Series([10, 20])
        standard_deviation_of_building_heights = pd.Series([2, 3])
        area_weighted_mean_of_building_heights = pd.Series([12, 22])
        building_surface_area_to_plan_area_ratio = pd.Series([1.2, 1.3])
        frontal_area_index = pd.DataFrame(
            {
                Settings.FRONTAL_AREA_INDEX_NORTH: [0.0, 0.1],
                Settings.FRONTAL_AREA_INDEX_EAST: [0.2, 0.3],
                Settings.FRONTAL_AREA_INDEX_SOUTH: [0.4, 0.5],
                Settings.FRONTAL_AREA_INDEX_WEST: [0.6, 0.7],
            }
        )
        complete_aspect_ratio = pd.Series([1.5, 1.6])
        height_to_width_ratio = pd.Series([1.7, 1.8])
        sky_view_factor = pd.Series([0.9, 0.8])
        grimmond_oke_roughness_length = pd.Series([0.7, 0.6])
        grimmond_oke_displacement_height = pd.Series([0.5, 0.4])
        raupach_roughness_length = pd.DataFrame(
            {
                Settings.RAUPACH_ROUGHNESS_LENGTH_NORTH: [0.3, 0.2],
                Settings.RAUPACH_ROUGHNESS_LENGTH_EAST: [0.2, 0.3],
                Settings.RAUPACH_ROUGHNESS_LENGTH_SOUTH: [0.4, 0.5],
                Settings.RAUPACH_ROUGHNESS_LENGTH_WEST: [0.6, 0.7],
            }
        )
        raupach_displacement_height = pd.DataFrame(
            {
                Settings.RAUPACH_DISPLACEMENT_HEIGHT_NORTH: [0.0, 0.1],
                Settings.RAUPACH_DISPLACEMENT_HEIGHT_EAST: [0.2, 0.3],
                Settings.RAUPACH_DISPLACEMENT_HEIGHT_SOUTH: [0.4, 0.5],
                Settings.RAUPACH_DISPLACEMENT_HEIGHT_WEST: [0.6, 0.7],
            }
        )
        macdonald_roughness_length = pd.DataFrame(
            {
                Settings.MACDONALD_ROUGHNESS_LENGTH_NORTH: [0.0, 0.1],
                Settings.MACDONALD_ROUGHNESS_LENGTH_EAST: [0.2, 0.3],
                Settings.MACDONALD_ROUGHNESS_LENGTH_SOUTH: [0.4, 0.5],
                Settings.MACDONALD_ROUGHNESS_LENGTH_WEST: [0.6, 0.7],
            }
        )
        macdonald_displacement_height = pd.Series([0.8, 0.9])
        vertical_distribution_of_building_heights = pd.DataFrame(
            {Settings.VERTICAL_DISTRIBUTION_OF_BUILDING_HEIGHTS: [10, 20]}
        )
        building_geometry = pd.Series([Point(0, 0), Point(1, 1)])
        target_crs = Settings.OUTPUT_CRS

        result = output.merge_parameters(
            frontal_area_density,
            plan_area_density,
            rooftop_area_density,
            plan_area_fraction,
            mean_building_height,
            standard_deviation_of_building_heights,
            area_weighted_mean_of_building_heights,
            building_surface_area_to_plan_area_ratio,
            frontal_area_index,
            complete_aspect_ratio,
            height_to_width_ratio,
            sky_view_factor,
            grimmond_oke_roughness_length,
            grimmond_oke_displacement_height,
            raupach_roughness_length,
            raupach_displacement_height,
            macdonald_roughness_length,
            macdonald_displacement_height,
            vertical_distribution_of_building_heights,
            building_geometry,
            target_crs,
        )

        assert isinstance(result, gpd.GeoDataFrame), "Output is not a GeoDataFrame"
        assert list(result.columns) == [
            "frontal_area_north",
            "frontal_area_east",
            "frontal_area_south",
            "frontal_area_west",
            "plan_area_density",
            "rooftop_area_density",
            "plan_area_fraction",
            "mean_building_height",
            "standard_deviation_of_building_heights",
            "area_weighted_mean_of_building_heights",
            "building_surface_area_to_plan_area_ratio",
            "frontal_area_index_north",
            "frontal_area_index_east",
            "frontal_area_index_south",
            "frontal_area_index_west",
            "complete_aspect_ratio",
            "height_to_width_ratio",
            "sky_view_factor",
            "grimmond_oke_roughness_length",
            "grimmond_oke_displacement_height",
            "raupach_roughness_length_north",
            "raupach_displacement_height_north",
            "raupach_roughness_length_east",
            "raupach_displacement_height_east",
            "raupach_roughness_length_south",
            "raupach_displacement_height_south",
            "raupach_roughness_length_west",
            "raupach_displacement_height_west",
            "macdonald_roughness_length_north",
            "macdonald_roughness_length_east",
            "macdonald_roughness_length_south",
            "macdonald_roughness_length_west",
            "macdonald_displacement_height",
            "vertical_distribution_of_building_heights",
            "building_geometry",
        ], "Output columns are not as expected"
        assert result.crs == target_crs, "Output CRS is not as expected"

    def test_numpy_to_binary(self):
        """Test the function `numpy_to_binary()` to ensure it outputs the right type and length binary file."""

        test_array = np.array([[[1, 2, 3], [4, 5, 6]], [[7, 8, 9], [10, 11, 12]]])

        binary_output = output.numpy_to_binary(test_array)

        assert isinstance(binary_output, bytes), "Output is not of type 'bytes'"
        assert len(binary_output) == 48, "Binary output length is not as expected"

    def test_raster_to_numpy(self):
        """Test the function `raster_to_numpy()` to ensure it outputs the right type and shape numpy array."""

        aggregate_rasters = xr.Dataset(
            {
                "parameter1": (("x", "y"), np.random.rand(5, 5)),
                "parameter2": (("x", "y"), np.random.rand(5, 5)),
                "building_count": (("x", "y"), np.random.randint(0, 10, (5, 5))),
            },
            coords={"x": range(5), "y": range(5)},
        )

        result = output.raster_to_numpy(aggregate_rasters)

        assert isinstance(result, np.ndarray), "Output is not a numpy array"
        assert result.shape == (132, 5, 5), "Output shape is not as expected"
        assert result.dtype == np.float32, "Output dtype is not np.float32"

    def test_rasterize_parameters(self):
        """Test the function `rasterize_parameters()` to ensure it outputs the right type and shape xr.Dataset."""

        merge_parameters = gpd.GeoDataFrame(
            {
                "parameter1": [1, 2, 3],
                "parameter2": [4, 5, 6],
                Settings.GEOMETRY_FIELD: [Point(0, 0), Point(1, 1), Point(2, 2)],
            }
        )

        result = output.rasterize_parameters(merge_parameters)

        assert isinstance(result, xr.Dataset), "Output is not an xr.Dataset"
        assert list(result.keys()) == [
            "parameter1",
            "parameter2",
            "building_count",
        ], "Output keys are not as expected"
        assert result["parameter1"].shape == (
            2400,
            2400,
        ), "Output shape for 'parameter1' is not as expected"
        assert result["parameter2"].shape == (
            2400,
            2400,
        ), "Output shape for 'parameter2' is not as expected"

    def test_write_binary(self):
        """Test that the function `write_binary()` writes a binary file correctly."""

        raster_to_numpy = np.zeros((132, 10, 10))
        test_binary_filename = "00001-00010.00001-00010"
        numpy_to_binary = raster_to_numpy.tobytes()

        output.write_binary(numpy_to_binary, raster_to_numpy)

        assert os.path.exists(test_binary_filename), "Binary file was not created."
        with open(test_binary_filename, "rb") as binary_file:
            content = binary_file.read()
            assert len(content) == 105648, "Content length is not as expected"

        os.remove(test_binary_filename)

    def test_write_index(self):
        """Test that the function `write_index()` writes an index file and contains the correct values."""

        raster_to_numpy = np.zeros((132, 10, 10))
        polygon1 = Polygon([[0, 0], [0, 1], [1, 1], [1, 0]])
        polygon2 = Polygon([[3, 3], [3, 4], [4, 4], [4, 3]])
        building_geometry = pd.Series([polygon1, polygon2])
        target_crs = "epsg:3857"
        test_index_filename = "test_index"

        file_path = self.test_dir + "/" + test_index_filename
        output.write_index(raster_to_numpy, building_geometry, target_crs, index_filename=file_path)

        assert os.path.exists(file_path), "Index file was not created."
        with open(file_path, "r") as index:
            content = index.read()
            assert "type=continuous" in content, "Index file type is not as expected."
            assert "projection=albers_nad83" in content, "Index file projection is not as expected."
            assert (
                "missing_value=-999900." in content
            ), "Index file missing_value is not as expected."
            assert "dy=0.00083333333" in content, "Index file dy is not as expected."
            assert "dx=0.00083333333" in content, "Index file dx is not as expected."
            assert "known_x=1" in content, "Index file known_x is not as expected."
            assert "known_y=1" in content, "Index file known_y is not as expected."
            assert "known_lat=0.0" in content, "Index file known_lat is not as expected."
            assert "known_lon=0.0" in content, "Index file known_lon is not as expected."
            assert "truelat1=45.5" in content, "Index file truelat1 is not as expected."
            assert "truelat2=29.5" in content, "Index file truelat2 is not as expected."
            assert (
                "stdlon=1.7966305682390428e-05" in content
            ), "Index file stdlon is not as expected."
            assert "wordsize=4" in content, "Index file wordsize is not as expected."
            assert "endian=big" in content, "Index file endian is not as expected."
            assert "signed=no" in content, "Index file signed is not as expected."
            assert "tile_x=10" in content, "Index file tile_x is not as expected."
            assert "tile_y=10" in content, "Index file tile_y is not as expected."
            assert "tile_z=132" in content, "Index file tile_z is not as expected."
            assert 'units="dimensionless"' in content, "Index file units is not as expected."
            assert "scale_factor=0.0001" in content, "Index file scale_factor is not as expected."
            assert (
                'description="Urban_Parameters"' in content
            ), "Index file description is not as expected."


if __name__ == "__main__":
    unittest.main()



import os
from typing import List, Union

import pandas as pd
from hamilton import driver, base
from hamilton.plugins import h_tqdm

import naturf.nodes as nodes
import naturf.output as output

DAGWORKS_API_KEY = os.environ.get("DAGWORKS_API_KEY")
DAGWORKS_USERNAME = os.environ.get("DAGWORKS_USERNAME")
DAGWORKS_PROJECT_ID = os.environ.get("DAGWORKS_PROJECT_ID")


class Model:
    def __init__(self, inputs: dict, outputs: List[str], **kwargs):
        # dictionary of parameter inputs required to construct the DAG
        self.inputs = inputs

        # desired output parameters
        self.outputs = outputs

        # instantiate any adapters we want
        hamilton_adapters = [
            base.SimplePythonDataFrameGraphAdapter(),
            h_tqdm.ProgressBar("Naturf DAG"),
        ]
        if DAGWORKS_API_KEY and DAGWORKS_USERNAME and DAGWORKS_PROJECT_ID:
            from dagworks import adapters

            hamilton_adapters.append(
                adapters.DAGWorksTracker(
                    project_id=int(DAGWORKS_PROJECT_ID),
                    api_key=DAGWORKS_API_KEY,
                    username=DAGWORKS_USERNAME,
                    dag_name="naturf-dag",
                    tags={"env": "dev", "status": "development", "version": "1"},
                )
            )

        # instantiate driver with function definitions & adapters
        self.dr = (
            driver.Builder()
            .with_config(self.inputs)
            .with_modules(nodes, output)
            .with_adapters(*hamilton_adapters)
            .build()
        )

    def execute(self) -> pd.DataFrame:
        """Run the driver."""

        # generate initial data frame
        df = self.dr.execute(self.outputs)

        return df

    def graph(self, view: bool = True, output_file_path: Union[str, None] = None) -> object:
        """Show the DAG. Return the graph object for the given inputs to execute."""

        return self.dr.visualize_execution(
            final_vars=self.outputs, output_file_path=output_file_path, render_kwargs={"view": view}
        )

    def list_parameters(self):
        """List all available parameters."""

        return self.dr.list_available_variables()



import re
from setuptools import setup, find_packages


def readme():
    """Return the contents of the project README file."""
    with open("README.md") as f:
        return f.read()


version = re.search(
    r"__version__ = ['\"]([^'\"]*)['\"]", open("naturf/__init__.py").read(), re.M
).group(1)

setup(
    name="naturf",
    version=version,
    packages=find_packages(),
    url="https://github.com/IMMM-SFA/naturf",
    license="MIT",
    author="Levi Sweet-Breu, Melissa Allen-Dumas, Chris R. Vernon, Emily Rexer",
    author_email="",
    description="An open-source Python package to address the effect of the geometry of a neighborhood on the local meteorology. ",
    long_description=readme(),
    long_description_content_type="text/markdown",
    python_requires=">=3.8.5",
    include_package_data=True,
    install_requires=[
        "numpy>=1.22.4",
        "pandas>=1.4.2",
        "geocube>=0.3.1",
        "rasterio>=1.2.10",
        "xarray>=2022.3.0",
        "joblib>=1.0.1",
        "fiona>=1.8.19",
        "pyproj>=3.0.1",
        "rtree>=1.0.0",
        "shapely>=1.8.2,<2",
        "geopandas>=0.10.2",
        "sf-hamilton[visualization]>=1.45",
        "tqdm",
    ],
)



import geopandas as gpd
import numpy as np
import os
import pandas as pd
from pyproj.crs import CRS
import struct
import xarray as xr

from functools import partial
from rasterio.enums import MergeAlg
from geocube.api.core import make_geocube
from geocube.rasterize import rasterize_image

from .config import Settings


def aggregate_rasters(rasterize_parameters: xr.Dataset) -> xr.Dataset:
    """Divide each raster by the number of buildings in the cell to get the average parameter value for each cell.

    :param rasterize_parameters:                    Xr.Dataset with rasterized parameters summed at the defined resolution.
    :type rasterize_parameters:                     Xr.Dataset

    :return:                                        Xr.Dataset containing rasterized parameter values averaged at the defined resolution.
    """

    return (rasterize_parameters / rasterize_parameters["building_count"]).fillna(0)


def merge_parameters(
    frontal_area_density: pd.DataFrame,
    plan_area_density: pd.DataFrame,
    rooftop_area_density: pd.DataFrame,
    plan_area_fraction: pd.Series,
    mean_building_height: pd.Series,
    standard_deviation_of_building_heights: pd.Series,
    area_weighted_mean_of_building_heights: pd.Series,
    building_surface_area_to_plan_area_ratio: pd.Series,
    frontal_area_index: pd.DataFrame,
    complete_aspect_ratio: pd.Series,
    height_to_width_ratio: pd.Series,
    sky_view_factor: pd.Series,
    grimmond_oke_roughness_length: pd.Series,
    grimmond_oke_displacement_height: pd.Series,
    raupach_roughness_length: pd.DataFrame,
    raupach_displacement_height: pd.DataFrame,
    macdonald_roughness_length: pd.DataFrame,
    macdonald_displacement_height: pd.Series,
    vertical_distribution_of_building_heights: pd.DataFrame,
    building_geometry: pd.Series,
    target_crs: CRS,
) -> gpd.GeoDataFrame:
    """Merge all parameters into one Pandas DataFrame.

    :param frontal_area_density:                       Frontal area density at each specified height increment and each cardinal direction.
    :type frontal_area_density:                        pd.DataFrame

    :param plan_area_density:                          Plan area density at each specified height increment.
    :type plan_area_density:                           pd.DataFrame

    :param rooftop_area_density:                       Rooftop area density at each specified height increment.
    :type rooftop_area_density:                        pd.DataFrame

    :param plan_area_fraction:                         Plan area fraction for each building.
    :type plan_area_fraction:                          pd.Series

    :param mean_building_height:                      Series of mean building height for all buildings within the target building's total plan area.
    :type mean_building_height:                       pd.Series

    :param standard_deviation_of_building_heights:     Series of standard deviation of building height for all buildings within the target building's
                                                       total plan area.
    :type standard_deviation_of_building_heights:      pd.Series

    :param area_weighted_mean_of_building_heights:     Series of area weighted mean building heights for each building.
    :type area_weighted_mean_of_building_heights:      pd.Series

    :param building_surface_area_to_plan_area_ratio:   Series of building surface area to plan area ratio for each building.
    :type building_surface_area_to_plan_area_ratio:    pd.Series

    :param frontal_area_index:                         Frontal area index for each building in each cardinal direction.
    :type frontal_area_index:                          pd.DataFrame

    :param complete_aspect_ratio:                      Complete aspect ratio for each building.
    :type complete_aspect_ratio:                       pd.Series

    :param height_to_width_ratio:                      Height-to-width ratio for each building.
    :type height_to_width_ratio:                       pd.Series

    :param sky_view_factor:                            Sky view factor for each building.
    :type sky_view_factor:                             pd.Series

    :param grimmond_oke_roughness_length:              Grimmond & Oke roughness length for each building.
    :type grimmond_oke_roughness_length:               pd.Series

    :param grimmond_oke_displacement_height:           Grimmond & Oke displacement height for each building.
    :type grimmond_oke_displacement_height:            pd.Series

    :param raupach_roughness_length:                   Raupach roughness length for each building in each cardinal direction.
    :type raupach_roughness_length:                    pd.DataFrame

    :param raupach_displacment_height:                 Raupach displacment height for each building in each cardinal direction.
    :type raupach_displacment_height:                  pd.DataFrame

    :param macdonald_roughness_length:                 Macdonald roughness_length for each building in each cardinal direction.
    :type macdonald_roughness_length:                  pd.DataFrame

    :param macdonald_displacement_height:              Macdonald displacement height for each building.
    :type macdonald_displacement_height:               pd.Series

    :param vertical_distribution_of_building_heights:  Distribution of building heights for each building ata each height increment.
    :type vertical_distribution_of_building_heights:   pd.DataFrame

    :param building_geometry:                          Geometry field for the buildings.
    :type building_geometry:                           pd.Series

    :param target_crs:                                 Coordinate reference system field of the parent geometry.
    :type target_crs:                                  CRS

    :return:                                           Pandas DataFrame with all parameters merged together.
    """

    df = pd.concat([frontal_area_density, plan_area_density, rooftop_area_density], axis=1)

    df[Settings.PLAN_AREA_FRACTION] = plan_area_fraction
    df[Settings.MEAN_BUILDING_HEIGHT] = mean_building_height
    df[Settings.STANDARD_DEVIATION_OF_BUILDING_HEIGHTS] = standard_deviation_of_building_heights
    df[Settings.AREA_WEIGHTED_MEAN_OF_BUILDING_HEIGHTS] = area_weighted_mean_of_building_heights
    df[Settings.BUILDING_SURFACE_AREA_TO_PLAN_AREA_RATIO] = building_surface_area_to_plan_area_ratio

    df = pd.concat([df, frontal_area_index], axis=1)

    df[Settings.COMPLETE_ASPECT_RATIO] = complete_aspect_ratio
    df[Settings.HEIGHT_TO_WIDTH_RATIO] = height_to_width_ratio
    df[Settings.SKY_VIEW_FACTOR] = sky_view_factor
    df[Settings.GRIMMOND_OKE_ROUGHNESS_LENGTH] = grimmond_oke_roughness_length
    df[Settings.GRIMMOND_OKE_DISPLACEMENT_HEIGHT] = grimmond_oke_displacement_height

    df[Settings.RAUPACH_ROUGHNESS_LENGTH_NORTH] = raupach_roughness_length[
        Settings.RAUPACH_ROUGHNESS_LENGTH_NORTH
    ]
    df[Settings.RAUPACH_DISPLACEMENT_HEIGHT_NORTH] = raupach_displacement_height[
        Settings.RAUPACH_DISPLACEMENT_HEIGHT_NORTH
    ]
    df[Settings.RAUPACH_ROUGHNESS_LENGTH_EAST] = raupach_roughness_length[
        Settings.RAUPACH_ROUGHNESS_LENGTH_EAST
    ]
    df[Settings.RAUPACH_DISPLACEMENT_HEIGHT_EAST] = raupach_displacement_height[
        Settings.RAUPACH_DISPLACEMENT_HEIGHT_EAST
    ]
    df[Settings.RAUPACH_ROUGHNESS_LENGTH_SOUTH] = raupach_roughness_length[
        Settings.RAUPACH_ROUGHNESS_LENGTH_SOUTH
    ]
    df[Settings.RAUPACH_DISPLACEMENT_HEIGHT_SOUTH] = raupach_displacement_height[
        Settings.RAUPACH_DISPLACEMENT_HEIGHT_SOUTH
    ]
    df[Settings.RAUPACH_ROUGHNESS_LENGTH_WEST] = raupach_roughness_length[
        Settings.RAUPACH_ROUGHNESS_LENGTH_WEST
    ]
    df[Settings.RAUPACH_DISPLACEMENT_HEIGHT_WEST] = raupach_displacement_height[
        Settings.RAUPACH_DISPLACEMENT_HEIGHT_WEST
    ]

    df = pd.concat([df, macdonald_roughness_length], axis=1)

    df[Settings.MACDONALD_DISPLACEMENT_HEIGHT] = macdonald_displacement_height

    df = pd.concat([df, vertical_distribution_of_building_heights], axis=1)

    df[Settings.GEOMETRY_FIELD] = building_geometry

    gdf = gpd.GeoDataFrame(df, geometry=Settings.GEOMETRY_FIELD, crs=target_crs)

    return gdf.to_crs(Settings.OUTPUT_CRS)


def numpy_to_binary(raster_to_numpy: np.ndarray) -> bytes:
    """Turn the master numpy array containing all 132 aggregated parameters into a binary stream.

    :param raster_to_numpy:         132 level numpy array with each level being an aggregated parameter.
    :type raster_to_numpy:          np.ndarray

    :return:                        Binary object containing the parameter data.
    """

    master_out = []

    for i in range(len(raster_to_numpy)):
        master_outi = bytes()
        for j in range(len(raster_to_numpy[i])):
            for k in range(len(raster_to_numpy[i][j])):
                master_outi += struct.pack(">i", int(raster_to_numpy[i][j][k]))
        master_out.append(master_outi)

    master_out_final = bytes()

    for i in range(len(master_out)):
        master_out_final += master_out[i]

    return master_out_final


def raster_to_numpy(aggregate_rasters: xr.Dataset) -> np.ndarray:
    """Stack all 132 rasterized parameters into one numpy array for conversion to a binary file.

    :param aggregate_rasters:                 Dataset with rasterized parameter values averaged at the defined resolution.
    :type aggregate_rasters:                  xr.Dataset

    :return:                                  132 level numpy array with each level being an aggregated parameter.
    """

    parameters = list(aggregate_rasters.keys())
    parameters.remove("building_count")
    rows = aggregate_rasters.dims["y"]
    cols = aggregate_rasters.dims["x"]
    master = np.zeros((132, rows, cols), dtype=np.float32)
    i = 0
    for parameter in parameters:
        master[i] = aggregate_rasters[parameter].to_numpy()
        i += 1

    return master * 10**Settings.SCALING_FACTOR


def rasterize_parameters(merge_parameters: gpd.GeoDataFrame) -> xr.Dataset:
    """Rasterize parameters in preparation for conversion to numpy arrays. Raster will be of resolution Settings.DEFAULT_OUTPUT_RESOLUTION
    and each cell will be the sum of each parameter value within. By default all_touched is True so that every building that is within a cell is
    included in the sum.

    :param merge_parameters:             Pandas.GeoDataFrame with all selected urban parameters for each building.
    :type merge_parameters:              Pandas.GeoDataFrame

    :return:                             Xr.Dataset containing rasterization of selected urban parameters.
    """

    merge_parameters["building_count"] = 1
    measurements = merge_parameters.columns[merge_parameters.columns != Settings.GEOMETRY_FIELD]
    resolution = Settings.DEFAULT_OUTPUT_RESOLUTION
    fill = Settings.DEFAULT_FILL_VALUE

    return make_geocube(
        vector_data=merge_parameters.rename(
            columns={Settings.GEOMETRY_FIELD: "geometry"}
        ).set_geometry("geometry"),
        measurements=measurements,
        resolution=resolution,
        fill=fill,
        rasterize_function=partial(rasterize_image, all_touched=True, merge_alg=MergeAlg.add),
    )


def write_index(
    raster_to_numpy: np.ndarray,
    building_geometry: pd.Series,
    target_crs: CRS,
    index_filename: str = "index",
) -> str:
    """Write the index file that will accompany the output binary file.

    :param raster_to_numpy:                 132 level numpy array with each level being an aggregated parameter.
    :type raster_to_numpy:                  np.ndarray

    :param building_geometry:               Geometry field for the buildings.
    :type building_geometry:                pd.Series

    :param target_crs:                      Coordinate reference system field of the parent geometry.
    :type target_crs:                       crs
    """

    dy = float(Settings.DEFAULT_OUTPUT_RESOLUTION[0])
    dx = float(Settings.DEFAULT_OUTPUT_RESOLUTION[1])

    building_geometry = gpd.GeoSeries(building_geometry, crs=target_crs)

    building_geometry_project = building_geometry.to_crs(crs=4326)

    bounds = building_geometry_project.total_bounds

    known_lat = bounds[1]
    known_lon = bounds[0]

    stdlon = (bounds[0] + bounds[2]) / 2

    tile_x = raster_to_numpy.shape[2]
    tile_y = raster_to_numpy.shape[1]

    scale_factor = 10**-Settings.SCALING_FACTOR

    with open(index_filename, "w") as index:
        index.writelines(
            [
                "type=continuous\n",
                "  projection=albers_nad83\n",
                "  missing_value=-999900.\n",
                "  dy=" + str(dy) + "\n",
                "  dx=" + str(dx) + "\n",
                "  known_x=1\n",
                "  known_y=1\n",
                "  known_lat=" + str(known_lat) + "\n",
                "  known_lon=" + str(known_lon) + "\n",
                "  truelat1=45.5\n",
                "  truelat2=29.5\n",
                "  stdlon=" + str(stdlon) + "\n",
                "  wordsize=4\n",
                "  endian=big\n",
                "  signed=no\n",
                "  tile_x=" + str(tile_x) + "\n",
                "  tile_y=" + str(tile_y) + "\n",
                "  tile_z=132\n",
                '  units="dimensionless"\n',
                "  scale_factor=" + str(scale_factor) + "\n",
                '  description="Urban_Parameters"\n',
            ]
        )


def write_binary(numpy_to_binary: bytes, raster_to_numpy: np.ndarray) -> None:
    """Write the binary file that will be input to WRF.

    :param numpy_to_binary:                 Binary object containing the parameter data.
    :type numpy_to_binary:                  bytes

    :param raster_to_numpy:                 132 level numpy array with each level being an aggregated parameter.
    :type raster_to_numpy:                  np.ndarray
    """

    np.save("temporary.npy", numpy_to_binary)

    rows = raster_to_numpy.shape[1]
    cols = raster_to_numpy.shape[2]

    first_y_index = "{:05d}".format(1)
    second_y_index = "{:05d}".format(rows)

    first_x_index = "{:05d}".format(1)
    second_x_index = "{:05d}".format(cols)

    out_binary_name = (
        first_x_index + "-" + second_x_index + "." + first_y_index + "-" + second_y_index
    )

    with open("temporary.npy", "rb") as tile, open(out_binary_name, "wb") as tile2:
        tile2.write(tile.read()[20 * 4 :])

    os.remove("temporary.npy")



class Settings:
    """
    This class is used to define variables for the naturf application.
    """

    BUILDING_HEIGHT_INTERVAL = 5
    VONKARMANCONSTANT = 0.4
    OBSTACLEDRAGCOEFFICIENT = 1.12
    PSI_H = 0.193
    DRAGCOEFFICIENT_03 = 0.3
    DRAGCOEFFICIENT_0003 = 0.003
    ALPHACOEFFICIENT = 3.59
    BETACOEFFICIENT = 1.0
    CAP_STYLE = 3
    CONSTANT_15 = 15
    CONSTANT_75 = 7.5
    DEGREES_IN_CIRCLE = 360
    DILAREA_DEFAULT = 10000
    DISPLACEMENT_HEIGHT_FACTOR = 0.67
    MAX_BUILDING_HEIGHT = 75
    NORTHEAST_DEGREES = 45
    NORTHWEST_DEGREES = 135
    OUTPUT_CRS = "EPSG:4326"
    RADIUS = 100
    RDH_THRESHOLD_MAX = 3
    RDH_THRESHOLD_MIN = 0
    ROUGHNESS_LENGTH_FACTOR = 0.1
    RRL_THRESHOLD_MAX = 3
    RRL_THRESHOLD_MIN = 0
    SMALL_DECIMAL = 0.0000000000000000000000000000001
    SOUTHEAST_DEGREES = 315
    SOUTHEAST_DEGREES_ARCTAN = -45
    SOUTHWEST_DEGREES = 225
    SOUTHWEST_DEGREES_ARCTAN = -135
    START_OF_CIRCLE_DEGREES = 0
    DEFAULT_STREET_WIDTH = 15
    DEFAULT_OUTPUT_RESOLUTION = [0.00083333333, 0.00083333333]
    DEFAULT_FILL_VALUE = 0
    SCALING_FACTOR = 4

    DATA_ID_FIELD_NAME = "OBJECTID"
    DATA_HEIGHT_FIELD_NAME = "Max_HOUSE_"
    DATA_GEOMETRY_FIELD_NAME = "geometry"
    TARGET = "target"
    NEIGHBOR = "neighbor"
    NORTH = "north"
    EAST = "east"
    SOUTH = "south"
    WEST = "west"
    NORTH_SOUTH = "north_south"
    EAST_WEST = "east_west"

    ID_FIELD = "building_id"
    HEIGHT_FIELD = "building_height"
    GEOMETRY_FIELD = "building_geometry"
    AREA_FIELD = "building_area"
    BUILDING_PLAN_AREA_FIELD = "building_plan_area"
    CENTROID_FIELD = "building_centroid"
    BUFFERED_FIELD = "building_buffered"
    TOTAL_PLAN_AREA = "total_plan_area"
    TOTAL_PLAN_AREA_GEOMETRY = "total_plan_area_geometry"
    VOLUME_FIELD = "volume_field"

    TARGET_HEIGHT_FIELD = f"{HEIGHT_FIELD}_{TARGET}"
    NEIGHBOR_HEIGHT_FIELD = f"{HEIGHT_FIELD}_{NEIGHBOR}"
    TARGET_ID_FIELD = f"{ID_FIELD}_{TARGET}"
    NEIGHBOR_ID_FIELD = f"{ID_FIELD}_{NEIGHBOR}"
    TARGET_AREA_FIELD = f"{AREA_FIELD}_{TARGET}"
    NEIGHBOR_AREA_FIELD = f"{AREA_FIELD}_{NEIGHBOR}"
    TARGET_GEOMETRY_FIELD = f"{GEOMETRY_FIELD}_{TARGET}"
    NEIGHBOR_GEOMETRY_FIELD = f"{GEOMETRY_FIELD}_{NEIGHBOR}"
    TARGET_CENTROID_FIELD = f"{CENTROID_FIELD}_{TARGET}"
    NEIGHBOR_CENTROID_FIELD = f"{CENTROID_FIELD}_{NEIGHBOR}"
    NEIGHBOR_VOLUME_FIELD = f"{VOLUME_FIELD}_{NEIGHBOR}"

    TARGET_BUFFERED_FIELD = f"{BUFFERED_FIELD}_{TARGET}"

    DISTANCE_TO_NEIGHBOR_BY_CENTROID = f"distance_to_{NEIGHBOR}_by_centroid"
    DISTANCE_BETWEEN_BUILDINGS = "distance_between_buildings"
    AVERAGE_DISTANCE_BETWEEN_BUILDINGS = "average_distance_between_buildings"

    SPATIAL_JOIN_LIST = [
        TARGET_HEIGHT_FIELD,
        NEIGHBOR_HEIGHT_FIELD,
        TARGET_ID_FIELD,
        NEIGHBOR_ID_FIELD,
        TARGET_AREA_FIELD,
        NEIGHBOR_AREA_FIELD,
        TARGET_GEOMETRY_FIELD,
        NEIGHBOR_GEOMETRY_FIELD,
    ]

    WALL_ANGLE = "wall_angle"
    WALL_DIRECTION = "wall_direction"
    WALL_LENGTH = "wall_length"
    WALL_LENGTH_NORTH = "wall_length_north"
    WALL_LENGTH_EAST = "wall_length_east"
    WALL_LENGTH_SOUTH = "wall_length_south"
    WALL_LENGTH_WEST = "wall_length_west"
    FRONTAL_LENGTH_NORTH = "frontal_length_north"
    FRONTAL_LENGTH_EAST = "frontal_length_east"
    FRONTAL_LENGTH_SOUTH = "frontal_length_south"
    FRONTAL_LENGTH_WEST = "frontal_length_west"
    FRONTAL_AREA_NORTH = "frontal_area_north"
    FRONTAL_AREA_EAST = "frontal_area_east"
    FRONTAL_AREA_SOUTH = "frontal_area_south"
    FRONTAL_AREA_WEST = "frontal_area_west"
    PLAN_AREA_DENSITY = "plan_area_density"
    ROOFTOP_AREA_DENSITY = "rooftop_area_density"
    PLAN_AREA_FRACTION = "plan_area_fraction"
    MEAN_BUILDING_HEIGHT = "mean_building_height"
    STANDARD_DEVIATION_OF_BUILDING_HEIGHTS = "standard_deviation_of_building_heights"
    AREA_WEIGHTED_MEAN_OF_BUILDING_HEIGHTS = "area_weighted_mean_of_building_heights"
    BUILDING_SURFACE_AREA = "building_surface_area"
    BUILDING_SURFACE_AREA_TO_PLAN_AREA_RATIO = "building_surface_area_to_plan_area_ratio"
    FRONTAL_AREA_INDEX_NORTH = "frontal_area_index_north"
    FRONTAL_AREA_INDEX_EAST = "frontal_area_index_east"
    FRONTAL_AREA_INDEX_SOUTH = "frontal_area_index_south"
    FRONTAL_AREA_INDEX_WEST = "frontal_area_index_west"
    COMPLETE_ASPECT_RATIO = "complete_aspect_ratio"
    HEIGHT_TO_WIDTH_RATIO = "height_to_width_ratio"
    SKY_VIEW_FACTOR = "sky_view_factor"
    GRIMMOND_OKE_ROUGHNESS_LENGTH = "grimmond_oke_roughness_length"
    GRIMMOND_OKE_DISPLACEMENT_HEIGHT = "grimmond_oke_displacement_height"
    RAUPACH_DISPLACEMENT_HEIGHT_NORTH = "raupach_displacement_height_north"
    RAUPACH_DISPLACEMENT_HEIGHT_EAST = "raupach_displacement_height_east"
    RAUPACH_DISPLACEMENT_HEIGHT_SOUTH = "raupach_displacement_height_south"
    RAUPACH_DISPLACEMENT_HEIGHT_WEST = "raupach_displacement_height_west"
    RAUPACH_ROUGHNESS_LENGTH_NORTH = "raupach_roughness_length_north"
    RAUPACH_ROUGHNESS_LENGTH_EAST = "raupach_roughness_length_east"
    RAUPACH_ROUGHNESS_LENGTH_SOUTH = "raupach_roughness_length_south"
    RAUPACH_ROUGHNESS_LENGTH_WEST = "raupach_roughness_length_west"
    MACDONALD_ROUGHNESS_LENGTH_NORTH = "macdonald_roughness_length_north"
    MACDONALD_ROUGHNESS_LENGTH_EAST = "macdonald_roughness_length_east"
    MACDONALD_ROUGHNESS_LENGTH_SOUTH = "macdonald_roughness_length_south"
    MACDONALD_ROUGHNESS_LENGTH_WEST = "macdonald_roughness_length_west"
    MACDONALD_DISPLACEMENT_HEIGHT = "macdonald_displacement_height"
    VERTICAL_DISTRIBUTION_OF_BUILDING_HEIGHTS = "vertical_distribution_of_building_heights"
